#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import date, timedelta
from typing import Mapping, Tuple, Type

import pytest

from nasty.request.replies import Replies
from nasty.request.request import Request
from nasty.request.search import Search
from nasty.request.thread import Thread


@pytest.mark.parametrize(
    "args",
    [
        (Search, {"query": "q", "max_tweets": 0}),
        (Search, {"query": "q", "max_tweets": -1}),
        (Search, {"query": "q", "batch_size": 0}),
        (Search, {"query": "q", "batch_size": -1}),
        (Search, {"query": "q", "since": date(2010, 1, 1), "until": date(2010, 1, 1)}),
        (Search, {"query": "q", "since": date(2010, 1, 2), "until": date(2010, 1, 1)}),
    ],
    ids=lambda args: args[0].__name__ + ": " + repr(args[1]),
)
def test_illegal_args(args: Tuple[Type[Request], Mapping[str, object]]) -> None:
    type_, kwargs = args
    with pytest.raises(ValueError):
        type_(**kwargs)


@pytest.mark.parametrize(
    "request_",
    [
        Search("q"),
        Replies("332308211321425920", max_tweets=None),
        Thread("332308211321425920", max_tweets=123, batch_size=456),
    ],
    ids=repr,
)
def test_json_conversion(request_: Request) -> None:
    assert request_ == request_.from_json(request_.to_json())


@pytest.mark.parametrize(
    "search",
    [Search("q", since=date(2010, 1, 1), until=date(2010, 2, 1))],
    ids=repr,
)
def test_search_to_daily_requests(search: Search) -> None:
    # assert is not None necessary for mypy type checking
    daily_requests = search.to_daily_requests()
    assert search.since is not None and search.until is not None
    assert (search.until - search.since).days == len(daily_requests)
    for daily_request in daily_requests:
        assert daily_request.since is not None and daily_request.until is not None
        assert timedelta(days=1) == daily_request.until - daily_request.since


@pytest.mark.parametrize(
    "search",
    [
        Search("q"),
        Search("q", since=date(2010, 1, 1)),
        Search("q", until=date(2010, 1, 1)),
    ],
    ids=repr,
)
def test_search_into_daily_requests_illegal_args(search: Search) -> None:
    with pytest.raises(ValueError):
        search.to_daily_requests()



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import json
from datetime import timedelta
from http.cookiejar import Cookie, DefaultCookiePolicy
from logging import getLogger
from pathlib import Path
from threading import Lock
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    cast,
)

from _pytest.monkeypatch import MonkeyPatch
from requests import PreparedRequest, Response, Session
from requests.cookies import RequestsCookieJar
from requests.hooks import default_hooks
from requests.structures import CaseInsensitiveDict
from typing_extensions import Final

logger = getLogger(__name__)
_SESSION_SEND: Final[Callable[..., Response]] = Session.send
_T_Serializable = TypeVar("_T_Serializable")


class Serializable(Generic[_T_Serializable]):
    def __init__(
        self,
        type_: Type[_T_Serializable],
        encoder: Callable[[_T_Serializable], Dict[str, object]],
        decoder: Optional[Callable[[Dict[str, object]], Dict[str, object]]] = None,
    ):
        self.type = type_
        self.encoder = encoder
        self.decoder = decoder


SERIALIZABLES: Mapping[str, Serializable[Any]] = {
    bytes.__name__: Serializable(
        type_=bytes,
        encoder=lambda obj: {"base64": base64.encodebytes(obj).decode("ascii")},
    ),
    timedelta.__name__: Serializable(
        type_=timedelta,
        encoder=lambda obj: {
            "days": obj.days,
            "seconds": obj.seconds,
            "microseconds": obj.microseconds,
        },
    ),
    CaseInsensitiveDict.__name__: Serializable(
        type_=CaseInsensitiveDict,
        encoder=lambda obj: cast(Dict[str, object], obj._store),
        decoder=lambda obj: {"_store": obj},
    ),
    Cookie.__name__: Serializable(
        type_=Cookie,
        encoder=lambda obj: cast(Dict[str, object], obj.__dict__),
        decoder=lambda obj: obj,
    ),
    RequestsCookieJar.__name__: Serializable(
        type_=RequestsCookieJar,
        encoder=lambda obj: dict(obj.__getstate__(), _policy=None),
        decoder=lambda obj: dict(obj, _policy=DefaultCookiePolicy()),
    ),
    PreparedRequest.__name__: Serializable(
        type_=PreparedRequest,
        encoder=lambda obj: dict(obj.__dict__, hooks=None),
        decoder=lambda obj: dict(
            obj, hooks=cast(Callable[[], Mapping[str, Sequence[Any]]], default_hooks)()
        ),
    ),
    Response.__name__: Serializable(
        type_=Response,
        encoder=lambda obj: dict(obj.__getstate__(), history=None),
        decoder=lambda obj: dict(obj, history=None),
    ),
}


def encode_json(obj: object) -> object:
    try:
        result = SERIALIZABLES[type(obj).__name__].encoder(obj)
    except KeyError:
        return obj
    result["__type__"] = type(obj).__name__
    return result


def decode_json(obj: Dict[str, object]) -> object:
    if not isinstance(obj, Dict) or "__type__" not in obj:
        return obj

    serializable = SERIALIZABLES[cast(str, obj.pop("__type__"))]
    if serializable.decoder is not None:
        result_dict = serializable.decoder(obj)
        result = object.__new__(serializable.type)
        if hasattr(result, "__setstate__"):
            result.__setstate__(result_dict)
        else:
            result.__dict__ = result_dict
        return result
    elif serializable.type == bytes:
        return base64.decodebytes(cast(str, obj["base64"]).encode("ascii"))
    elif serializable.type == timedelta:
        return timedelta(
            days=cast(float, obj["days"]),
            seconds=cast(float, obj["seconds"]),
            microseconds=cast(float, obj["microseconds"]),
        )
    raise ValueError(
        "Decoding type {} not implemented.".format(serializable.type.__name__)
    )


def _request_to_cache_key(request: PreparedRequest) -> Tuple[Any, ...]:
    return (
        request.method,
        request.url,
        tuple(request.headers.items()),
        tuple(request._cookies.items()),  # type: ignore
        request.body,
        request._body_position,  # type: ignore
    )


class RequestsCache:
    def __init__(self) -> None:
        self._lock = Lock()
        self._cache: Dict[Tuple[Any, ...], Response] = {}
        self._cache_file = Path(__file__).parent / ".requests_cache.jsonl"
        if self._cache_file.exists():
            with self._cache_file.open("rt", encoding="UTF-8") as fin:
                for line in fin:
                    response = json.loads(line, object_hook=decode_json)
                    self._cache[_request_to_cache_key(response.request)] = response

    def close(self) -> None:
        with self._cache_file.open("wt", encoding="UTF-8") as fout:
            for response in self._cache.values():
                fout.write(json.dumps(response, default=encode_json))
                fout.write("\n")

    def __enter__(self) -> "RequestsCache":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def activate(self, monkeypatch: MonkeyPatch, regenerate: bool = False) -> None:
        def mock_session_send(
            session: Session, request: PreparedRequest, **kwargs: Any
        ) -> Response:
            with self._lock:
                key = _request_to_cache_key(request)
                response = self._cache.get(key)
                if response is not None:
                    logger.debug(
                        "Found cache response"
                        + ("." if not regenerate else " (regenerating).")
                    )

                    if not regenerate:
                        session.cookies.update(response.cookies)  # type: ignore
                        return response

                    self._cache.pop(key)

                response = _SESSION_SEND(session, request, **kwargs)
                self._cache[key] = response

                return response

        monkeypatch.setattr(Session, Session.send.__name__, mock_session_send)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from argparse import ArgumentTypeError
from datetime import date, datetime, timedelta
from typing import Iterable


# Adapted from https://stackoverflow.com/a/25470943/211404
def yyyy_mm_dd_date(string: str) -> date:
    try:
        return datetime.strptime(string, "%Y-%m-%d").date()
    except ValueError:
        raise ArgumentTypeError(
            'Can not parse date: "{}". Make sure it is in '
            "YYYY-MM-DD format.".format(string)
        )


# Adapted from: https://stackoverflow.com/a/1060352/211404
def daterange(start_date: date, end_date: date) -> Iterable[date]:
    if start_date > end_date:
        raise ValueError(
            "End date {} before start date {}.".format(start_date, end_date)
        )

    current_date = start_date
    delta = timedelta(days=1)
    while current_date <= end_date:
        yield current_date
        current_date += delta



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from logging import getLogger

import pytest
from _pytest.capture import CaptureFixture

from nasty import main

logger = getLogger(__name__)


@pytest.mark.parametrize(
    "args_string",
    [
        "trump",
        "332308211321425920",
        "search trump",
        "search --query trump --since 2019",
        "search --query trump --until 2019",
        "search --query trump --filter latest",
        "search --query trump --max-tweets five",
        "search --query trump --batch-size 3.0",
        "search --query trump --to-batch",
        "search --query trump --daily",
        "search --query trump --to-batch file --daily",
        "search --query trump --since 2019-03-21 --to-batch file --daily",
        "search --query trump --until 2019-03-21 --to-batch file --daily",
        "replies 332308211321425920",
        "replies --tweet-id 332308211321425920 --max-tweets five",
        "replies --tweet-id 332308211321425920 --batch-size 3.0",
        "replies --tweet-id 332308211321425920 --to-batch",
        "thread 332308211321425920",
        "thread --tweet-id 332308211321425920 --max-tweets five",
        "thread --tweet-id 332308211321425920 --batch-size 3.0",
        "thread --tweet-id 332308211321425920 --to-batch",
        "batch",
        "batch --batch-file",
        "batch --batch-file batch.jsonl",
        "batch --batch-file batch.jsonl --results-dir",
        "batch --results-dir",
        "batch --batch-file --results-dir out/",
        "idify --in-dir",
        "idify --out-dir",
        "idify --out-dir out/",
        "unidify --in-dir",
        "unidify --out-dir",
        "unidify --out-dir out/",
    ],
    ids=repr,
)
def test_incorrect_calls(args_string: str, capsys: CaptureFixture) -> None:
    args = args_string.split(" ") if args_string != "" else []
    logger.debug("Raw arguments: {}".format(args))

    with pytest.raises(SystemExit) as excinfo:
        main(*args)

    assert excinfo.value.code == 2

    captured = capsys.readouterr().err
    logger.debug("Captured Error:")
    for line in captured.split("\n"):
        logger.debug("  " + line)
    assert "usage: nasty" in captured and ": error: " in captured



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import lzma
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Iterator, TextIO, cast


@contextmanager
def _read_file(file: Path, *, use_lzma: bool = False) -> Iterator[TextIO]:
    if use_lzma:
        with lzma.open(file, "rt", encoding="UTF-8") as fin:
            yield cast(TextIO, fin)
    else:
        with file.open("r", encoding="UTF-8") as fin:
            yield fin


@contextmanager
def _write_file_with_tmp_guard(
    file: Path, *, overwrite_existing: bool = False, use_lzma: bool = False
) -> Iterator[TextIO]:
    if not overwrite_existing and file.exists():
        raise ValueError(
            "File '{}' to be written to, already exists. Manual intervention required! "
            "Check file and delete if no longer needed.".format(file)
        )

    tmp_file = file.parent / (".tmp." + file.name)

    if use_lzma:
        with lzma.open(tmp_file, "wt", encoding="UTF-8") as fin:
            yield cast(TextIO, fin)
    else:
        with tmp_file.open("w", encoding="UTF-8") as fin:
            yield fin

    tmp_file.rename(file)


def read_file(file: Path, *, use_lzma: bool = False) -> str:
    with _read_file(file, use_lzma=use_lzma) as fin:
        return fin.read()


def write_file(
    file: Path, value: str, *, overwrite_existing: bool = False, use_lzma: bool = False
) -> None:
    with _write_file_with_tmp_guard(
        file, overwrite_existing=overwrite_existing, use_lzma=use_lzma
    ) as fout:
        fout.write(value)


def read_lines_file(file: Path, *, use_lzma: bool = False) -> Iterable[str]:
    with _read_file(file, use_lzma=use_lzma) as fin:
        for line in fin:
            yield line.strip()


def write_lines_file(
    file: Path,
    values: Iterable[str],
    *,
    overwrite_existing: bool = False,
    use_lzma: bool = False,
) -> None:
    with _write_file_with_tmp_guard(
        file, overwrite_existing=overwrite_existing, use_lzma=use_lzma
    ) as fout:
        for value in values:
            fout.write(value)
            fout.write("\n")



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Iterator

import pytest
from _pytest.config import Config
from _pytest.fixtures import FixtureRequest
from _pytest.monkeypatch import MonkeyPatch
from nasty_utils import LoggingSettings

from nasty._settings import NastySettings

from .util.requests_cache import RequestsCache


def pytest_configure(config: Config) -> None:
    LoggingSettings.setup_pytest_logging(config)
    _configure_requests_cache(config)


def _configure_requests_cache(config: Config) -> None:
    config.addinivalue_line(
        "markers", "requests_cache_disabled: Disable caching of requests."
    )
    config.addinivalue_line(
        "markers", "requests_cache_regenerate: Regenerate requested cached requests."
    )


@pytest.fixture(scope="session")
def requests_cache() -> Iterator[RequestsCache]:
    with RequestsCache() as requests_cache:
        yield requests_cache


@pytest.fixture(scope="session")
def settings() -> Iterator[NastySettings]:
    settings = NastySettings.find_and_load_from_settings_file()
    yield settings


@pytest.fixture(autouse=True)
def activate_requests_cache(
    request: FixtureRequest, monkeypatch: MonkeyPatch, requests_cache: RequestsCache
) -> None:
    if not request.node.get_closest_marker("requests_cache_disabled"):
        requests_cache.activate(
            monkeypatch,
            bool(request.node.get_closest_marker("requests_cache_regenerate")),
        )


@pytest.fixture(autouse=True)
def disrespect_robotstxt(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("NASTY_DISRESPECT_ROBOTSTXT", "1")



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import datetime, timezone

from nasty.tweet.tweet import Tweet, User

tweet_jsons = {
    # Tweet accessed via Search API on 2019-09-27
    "1142944425502543875": {
        "created_at": "Sun Jun 23 23:56:00 +0000 2019",
        "id": 1142944425502543875,
        "id_str": "1142944425502543875",
        "full_text": (
            "It is obvious that Mr. Trump is a terrible judge of character. His "
            "nominees for various high level cabinet positions have been unqualified, "
            "unforthcoming, or engaged in behavior unbecoming of a public servant. His "
            "standards should not be our standards. https://t.co/CHuQqFs5EX"
        ),
        "truncated": False,
        "display_text_range": [0, 276],
        "entities": {
            "hashtags": [],
            "symbols": [],
            "user_mentions": [],
            "urls": [
                {
                    "url": "https://t.co/CHuQqFs5EX",
                    "expanded_url": (
                        "https://www.washingtonpost.com/opinions/global-opinions/the-wh"
                        "ite-house-is-allowing-personal-scandals-to-become-national-emb"
                        "arrassments/2019/06/20/122f4e84-938b-11e9-b570-6416efdc0803_st"
                        "ory.html?utm_term=.790cde3fee18"
                    ),
                    "display_url": "washingtonpost.com/opinions/globa…",
                    "indices": [253, 276],
                }
            ],
        },
        "source": (
            '<a href="https://sproutsocial.com" rel="nofollow">Sprout Social</a>"'
        ),
        "in_reply_to_status_id": None,
        "in_reply_to_status_id_str": None,
        "in_reply_to_user_id": None,
        "in_reply_to_user_id_str": None,
        "in_reply_to_screen_name": None,
        "geo": None,
        "coordinates": None,
        "place": None,
        "contributors": None,
        "is_quote_status": False,
        "retweet_count": 1109,
        "favorite_count": 3709,
        "reply_count": 241,
        "conversation_id": 1142944425502543875,
        "conversation_id_str": "1142944425502543875",
        "favorited": False,
        "retweeted": False,
        "possibly_sensitive": False,
        "possibly_sensitive_editable": True,
        "lang": "en",
        "supplemental_language": None,
        "user": {
            "id": 949934436,
            "id_str": "949934436",
            "name": "Tom Steyer",
            "screen_name": "TomSteyer",
            "location": "San Francisco, CA",
            "description": (
                "Husband & father. Former investor. Signed The Giving Pledge to donate "
                "the majority of my net worth in my lifetime. He/him. 2020 Democrat "
                "for president."
            ),
            "url": "https://t.co/n1r4Xd6IXH",
            "entities": {
                "url": {
                    "urls": [
                        {
                            "url": "https://t.co/n1r4Xd6IXH",
                            "expanded_url": "https://go.tomsteyer.com/ts2020",
                            "display_url": "go.tomsteyer.com/ts2020",
                            "indices": [0, 23],
                        }
                    ]
                },
                "description": {"urls": []},
            },
            "protected": False,
            "followers_count": 245080,
            "fast_followers_count": 0,
            "normal_followers_count": 245080,
            "friends_count": 1086,
            "listed_count": 2764,
            "created_at": "Thu Nov 15 15:28:50 +0000 2012",
            "favourites_count": 3535,
            "utc_offset": None,
            "time_zone": None,
            "geo_enabled": False,
            "verified": True,
            "statuses_count": 8513,
            "media_count": 812,
            "lang": None,
            "contributors_enabled": False,
            "is_translator": False,
            "is_translation_enabled": False,
            "profile_background_color": "666666",
            "profile_background_image_url": (
                "http://abs.twimg.com/images/themes/theme2/bg.gif"
            ),
            "profile_background_image_url_https": (
                "https://abs.twimg.com/images/themes/theme2/bg.gif"
            ),
            "profile_background_tile": False,
            "profile_image_url": (
                "http://pbs.twimg.com/profile_images/1148571376355119104/"
                "EqZutNij_normal.png"
            ),
            "profile_image_url_https": (
                "https://pbs.twimg.com/profile_images/1148571376355119104/"
                "EqZutNij_normal.png"
            ),
            "profile_banner_url": (
                "https://pbs.twimg.com/profile_banners/949934436/1562688733"
            ),
            "profile_image_extensions_media_color": {
                "palette": [
                    {"rgb": {"red": 23, "green": 44, "blue": 73}, "percentage": 50.97},
                    {
                        "rgb": {"red": 202, "green": 158, "blue": 142},
                        "percentage": 29.1,
                    },
                    {"rgb": {"red": 112, "green": 77, "blue": 69}, "percentage": 11.04},
                    {
                        "rgb": {"red": 213, "green": 224, "blue": 232},
                        "percentage": 7.46,
                    },
                    {"rgb": {"red": 18, "green": 39, "blue": 69}, "percentage": 0.3},
                ]
            },
            "profile_image_extensions_alt_text": None,
            "profile_image_extensions_media_availability": None,
            "profile_image_extensions": {
                "mediaStats": {"r": {"missing": None}, "ttl": -1}
            },
            "profile_banner_extensions_media_color": {
                "palette": [
                    {
                        "rgb": {"red": 179, "green": 154, "blue": 132},
                        "percentage": 28.83,
                    },
                    {"rgb": {"red": 71, "green": 48, "blue": 42}, "percentage": 28.18},
                    {
                        "rgb": {"red": 174, "green": 186, "blue": 222},
                        "percentage": 17.33,
                    },
                    {
                        "rgb": {"red": 115, "green": 112, "blue": 117},
                        "percentage": 11.66,
                    },
                    {"rgb": {"red": 155, "green": 92, "blue": 87}, "percentage": 1.96},
                ]
            },
            "profile_banner_extensions_alt_text": None,
            "profile_banner_extensions_media_availability": None,
            "profile_banner_extensions": {
                "mediaStats": {"r": {"missing": None}, "ttl": -1}
            },
            "profile_link_color": "2802A3",
            "profile_sidebar_border_color": "000000",
            "profile_sidebar_fill_color": "FFEBD6",
            "profile_text_color": "605CBA",
            "profile_use_background_image": True,
            "has_extended_profile": False,
            "default_profile": False,
            "default_profile_image": False,
            "pinned_tweet_ids": [1175916938305949696],
            "pinned_tweet_ids_str": ["1175916938305949696"],
            "has_custom_timelines": True,
            "can_dm": None,
            "can_media_tag": None,
            "following": None,
            "follow_request_sent": None,
            "notifications": None,
            "muting": None,
            "blocking": None,
            "blocked_by": None,
            "want_retweets": None,
            "advertiser_account_type": "promotable_user",
            "advertiser_account_service_levels": [
                "dso",
                "media_studio",
                "smb",
                "dso",
                "mms",
                "smb",
                "dso",
            ],
            "profile_interstitial_type": "",
            "business_profile_state": "none",
            "translator_type": "none",
            "followed_by": None,
            "require_some_consent": False,
        },
    }
}


def test_1142944425502543875() -> None:
    tweet = Tweet(tweet_jsons["1142944425502543875"])
    assert (
        datetime(
            year=2019,
            month=6,
            day=23,
            hour=23,
            minute=56,
            second=0,
            tzinfo=timezone.utc,
        )
        == tweet.created_at
    )
    assert "1142944425502543875" == tweet.id
    assert (
        "It is obvious that Mr. Trump is a terrible judge of character. His nominees "
        "for various high level cabinet positions have been unqualified, "
        "unforthcoming, or engaged in behavior unbecoming of a public servant. His "
        "standards should not be our standards. https://t.co/CHuQqFs5EX" == tweet.text
    )
    assert "https://twitter.com/TomSteyer/status/1142944425502543875" == tweet.url
    assert tweet == Tweet.from_json(tweet.to_json())

    user = tweet.user
    assert "949934436" == user.id
    assert "Tom Steyer" == user.name
    assert "TomSteyer" == user.screen_name
    assert "https://twitter.com/TomSteyer" == user.url
    assert user == User.from_json(user.to_json())



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse

from overrides import overrides


# Adapted from: https://stackoverflow.com/a/9643162/211404
class SingleMetavarHelpFormatter(argparse.HelpFormatter):
    @overrides
    def _format_action_invocation(self, action: argparse.Action) -> str:
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            (metavar,) = self._metavar_formatter(action, default)(1)
            return metavar

        result = ", ".join(action.option_strings)
        if action.nargs != 0:
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            result += " " + args_string
        return result



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from abc import ABC, abstractmethod
from http import HTTPStatus
from logging import getLogger
from os import getenv
from time import sleep
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    cast,
)

import requests
from overrides import overrides
from requests.adapters import HTTPAdapter
from typing_extensions import Final, final
from urllib3 import Retry

from .._util.errors import UnexpectedStatusCodeException
from .._util.typing_ import checked_cast
from ..request.request import Request
from ..tweet.tweet import Tweet, TweetId, UserId
from ..tweet.tweet_stream import TweetStream

logger = getLogger(__name__)

crawl_delay: Optional[float] = None


class RetrieverTweetStream(TweetStream):
    def __init__(self, update_callback: Callable[[], bool]):
        self._update_callback: Final = update_callback
        self._tweets: Sequence[Tweet] = []
        self._tweets_position = 0

    def update_tweets(self, tweets: Sequence[Tweet]) -> None:
        self._tweets = tweets
        self._tweets_position = 0

    @overrides
    def __next__(self) -> Tweet:
        if self._tweets_position == len(self._tweets):
            if not self._update_callback():
                raise StopIteration()

        self._tweets_position += 1
        return self._tweets[self._tweets_position - 1]


class RetrieverBatch(ABC):
    def __init__(self, json: Mapping[str, Mapping[str, object]]):
        self._json: Final = json
        self.tweets: Final = self._tweets()
        self.next_cursor: Final = self._next_cursor()

    @final
    def _tweets(self) -> Sequence[Tweet]:
        id_to_tweet_json: Final = cast(
            Mapping[TweetId, Mapping[str, object]],
            self._json["globalObjects"]["tweets"],
        )
        id_to_user_json: Final = cast(
            Mapping[UserId, object], self._json["globalObjects"]["users"]
        )

        result = []
        for tweet_id in self._tweet_ids():
            if tweet_id not in id_to_tweet_json:
                # For conversation it can sometimes happen that a Tweet-ID is returned
                # without accompanying meta information. I have no idea why this happens
                # or how to fix it.
                logger.warning(
                    "Found Tweet-ID {} in timeline, but did not receive "
                    "Tweet meta information.".format(tweet_id)
                )
                # TODO: move this to a ConversationRetrieverBatch
                # TODO: add way to expose this over api
                continue

            tweet_json = dict(id_to_tweet_json[tweet_id])
            tweet_json["user"] = id_to_user_json[
                checked_cast(UserId, tweet_json["user_id_str"])
            ]

            # Delete remaining user fields in order to be similar to the Twitter
            # developer API and because the information is stored in the user object
            # anyways.
            tweet_json.pop("user_id", None)  # present on Search, not on Conversation
            tweet_json.pop("user_id_str")

            result.append(Tweet(tweet_json))
        return result

    @abstractmethod
    def _tweet_ids(self) -> Iterable[TweetId]:
        raise NotImplementedError()

    @abstractmethod
    def _next_cursor(self) -> Optional[str]:
        raise NotImplementedError()


_T_Request = TypeVar("_T_Request", bound=Request)


class Retriever(Generic[_T_Request], ABC):
    """Retrieves Tweets belonging to a specific Twitter timeline view.

    Implemented via Twitter's mobile web interface. For this we emulate what a normal
    browser would do:
    1) Load the HTML stub belonging to a timeline page. We say stub here, because the
       HTML doesn't contained any contents, i.e., there are no Tweets in it.
    2) Load batches of displayed Tweets via AJAX requests on page load and whenever the
       user scrolls to the bottom of the page.
    The upside of this approach is that the JSON results have the exact same format as
    the results from the Twitter developer API (and even contain more information).
    """

    def __init__(self, request: _T_Request):
        self._tweet_stream: Final = self._tweet_stream_type()(self._update_tweet_stream)
        self._request: Final = request
        self._session: Final = requests.Session()
        self._request_finished = False
        self._retrieved_tweets = 0
        self._cursor: Optional[str] = None

        # Configure on which status codes we should perform automated retries.
        self._session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(
                    total=5,
                    connect=5,
                    redirect=10,
                    backoff_factor=0.1,
                    raise_on_redirect=True,
                    raise_on_status=True,
                    status_forcelist=[
                        HTTPStatus.REQUEST_TIMEOUT,  # HTTP 408
                        HTTPStatus.CONFLICT,  # HTTP 409
                        HTTPStatus.INTERNAL_SERVER_ERROR,  # HTTP 500
                        HTTPStatus.NOT_IMPLEMENTED,  # HTTP 501
                        HTTPStatus.BAD_GATEWAY,  # HTTP 502
                        HTTPStatus.SERVICE_UNAVAILABLE,  # HTTP 503
                        HTTPStatus.GATEWAY_TIMEOUT,  # HTTP 504
                    ],
                )
            ),
        )

        self._fetch_new_twitter_session()

    @classmethod
    def _tweet_stream_type(cls) -> Type[RetrieverTweetStream]:
        return RetrieverTweetStream

    @classmethod
    @abstractmethod
    def _retriever_batch_type(cls) -> Type[RetrieverBatch]:
        raise NotImplementedError()

    @property
    def tweet_stream(self) -> RetrieverTweetStream:
        return self._tweet_stream

    @abstractmethod
    def _timeline_url(self) -> Mapping[str, object]:
        raise NotImplementedError()

    @abstractmethod
    def _batch_url(self) -> Mapping[str, object]:
        raise NotImplementedError()

    def _update_tweet_stream(self) -> bool:  # noqa: C901
        # TODO: try to reduce complexity and get red of noqa

        if self._request_finished:
            return False

        consecutive_rate_limits = 0
        consecutive_forbidden = 0
        consecutive_empty_batches = 0
        batch = None
        while True:
            try:
                batch = self._fetch_batch()
            except UnexpectedStatusCodeException as e:
                if e.status_code == HTTPStatus.TOO_MANY_REQUESTS:  # HTTP 429
                    consecutive_rate_limits += 1
                    if consecutive_rate_limits != 3:
                        self._fetch_new_twitter_session()
                        continue
                    logger.info("Received 3 consecutive TOO MANY REQUESTS responses.")
                elif e.status_code == HTTPStatus.FORBIDDEN:  # HTTP 403
                    consecutive_forbidden += 1
                    if consecutive_forbidden != 3:
                        self._fetch_new_twitter_session()
                        continue
                    logger.info("Received 3 consecutive FORBIDDEN responses.")
                raise
            consecutive_rate_limits = 0

            # Stop the iteration once the returned batch no longer contains any Tweets.
            # Ideally, we would like to omit this last request but there seems to be no
            # way to detect this prior to having the last batch loaded. Additionally,
            # Twitter will sometimes stop sending results early, which we also can not
            # detect. Because of this, we only stop loading once we receive empty
            # batches multiple times in a row.
            if not batch.tweets:
                consecutive_empty_batches += 1
                if consecutive_empty_batches != 3:
                    continue
                logger.info("Received 3 consecutive empty batches.")
                return False

            break

        tweets = batch.tweets
        if self._request.max_tweets:
            tweets = tweets[: self._request.max_tweets - self._retrieved_tweets]
        self._retrieved_tweets += len(tweets)
        logger.debug(
            "  Received new batch of {} Tweets ({}/{})".format(
                len(tweets), self._retrieved_tweets, self._request.max_tweets
            )
        )
        if (
            self._request.max_tweets
            and self._request.max_tweets == self._retrieved_tweets
        ):
            self._request_finished = True
        self.tweet_stream.update_tweets(tweets)

        self._cursor = batch.next_cursor
        if self._cursor is None:
            self._request_finished = True
        return True

    @final
    def _fetch_new_twitter_session(self) -> None:
        """Establishes a session with Twitter, so that they answer our requests.

        If we try to directly access request the first batch of a query, Twitter will
        respond with a rate limit error, i.e. HTTP 429. To receive actual responses we
        need to include a bearer token and a guest token in our headers. A normal web
        browser gets these be first loading the displaying HTML stub. This function
        emulates this process and prepares the given session object to contain the
        necessary headers.

        For more information on this process, see:
        - https://tech.b48.club/2019/05/13/how-to-fake-a-source-of-a-tweet.html
        - https://steemit.com/technology/@singhpratyush/
          fetching-url-for-complete-twitter-videos-using-guest-user-access-pattern
        - https://github.com/ytdl-org/youtube-dl/issues/12726#issuecomment-304779835

        Each established session is only good for a given number of requests.
        Information on this can be obtained by checking the X-Rate-Limit-* headers in
        the responses from api.twitter.com. Currently, we do not pay attention to these,
        and just establish a new session once we run into the first rate limit error.
        Cursor parameters, i.e. those that specify the current position in the result
        list seem to persist across sessions.

        Technically, a normal web browser would also receive a few cookies from Twitter
        in this process. Currently, api.twitter.com doesn't seem to check for these. In
        any case, we still set those in case Twitter changes their behavior. Note,
        however, that our requests will still be trivially distinguishable from a
        normal web browsers requests, as they typically sent many more headers and
        cookies, i.e. those from Google Analytics. Further we include the string
        "NASTYbot" in our User-Agent header to make it trivial for Twitter to
        rate-limit us, should they decide to.
        """

        logger.debug("  Establishing new Twitter session.")

        self._session.headers.clear()
        self._session.cookies.clear()

        # We use the current Chrome User-Agent string to get the most recent version of
        # the Twitter mobile website.
        self._session.headers["User-Agent"] = (
            "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N)"
            " AppleWebKit/537.36 (KHTML, like Gecko)"
            " Chrome/68.0.3440.84 Mobile Safari/537.36"
            " NASTYbot"
        )

        # The following header should not matter for the actual returned Tweets. Still,
        # since api.twitter.com also returns some localized strings for the UI (e.g.
        # headings), we set this to English, so these strings are always the same. If
        # not set, Twitter will guesstimate the language from the IP.
        self._session.headers["Accept-Language"] = "en_US,en"

        # Query HTML stub page. Also automatically adds any returned cookies by Twitter
        # via response headers to the session.
        response = self._session_get(**self._timeline_url())
        main_js_url = re.findall(
            "(https://abs.twimg.com/responsive-web/"
            "(?:client[-_])?web(?:[-_]legacy)?/main.[a-z0-9]+.js)",
            response.text,
        )[0]
        guest_token = re.findall(
            'document\\.cookie = decodeURIComponent\\(\\"gt=([0-9]+);', response.text
        )[0]

        # Queries the JS-script that carries the bearer token. Currently, this does not
        # seem to constant for all users, but we still check in case this changes in the
        # future.
        response = self._session_get(main_js_url)
        bearer_token = re.findall('.="Web-12",.="([^"]+)"', response.text)[0]

        # Emulate cookie setting that would be performed via Javascript.
        self._session.cookies.set_cookie(  # type: ignore
            requests.cookies.create_cookie(
                "gt", guest_token, domain=".twitter.com", path="/"
            )
        )

        # Set the two headers that we need to access api.twitter.com.
        self._session.headers["Authorization"] = "Bearer {}".format(bearer_token)
        self._session.headers["X-Guest-Token"] = guest_token

        logger.debug(
            "    Guest token: {}. Bearer token: {}.".format(guest_token, bearer_token)
        )

    @final
    def _fetch_batch(self) -> RetrieverBatch:
        return self._retriever_batch_type()(
            self._session_get(**self._batch_url()).json()
        )

    @final
    def _session_get(self, url: str, **kwargs: Any) -> requests.Response:
        if not getenv("NASTY_DISRESPECT_ROBOTSTXT"):
            global crawl_delay
            if crawl_delay is None:
                response = self._session.get("https://mobile.twitter.com/robots.txt")

                for line in response.text.splitlines():
                    if line.lower().startswith("crawl-delay:"):
                        crawl_delay = float(line[len("crawl-delay:") :])
                        break
                else:
                    raise RuntimeError("Could not determine crawl-delay.")

                logger.debug(
                    "    Determined crawl-delay of {:.2f}s.".format(crawl_delay)
                )

            sleep(crawl_delay)

        response = self._session.get(url, **kwargs)

        status = HTTPStatus(response.status_code)
        logger.debug(
            "    Received {} {} for {}".format(status.value, status.name, response.url)
        )
        if response.status_code != HTTPStatus.OK.value:
            raise UnexpectedStatusCodeException(
                response.url, HTTPStatus(response.status_code)
            )

        return response



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from http import HTTPStatus
from logging import getLogger
from timeit import timeit

import pytest
from requests import Session

logger = getLogger(__name__)
time_without_cache = None
time_with_cache = None


def _perform_request() -> None:
    with Session() as session:
        response = session.get("https://example.org")
    assert HTTPStatus(response.status_code) == HTTPStatus.OK


@pytest.mark.requests_cache_regenerate
def test_request_without_cache() -> None:
    global time_without_cache
    time_without_cache = timeit(_perform_request, number=1)


def test_request_with_cache() -> None:
    global time_with_cache
    time_with_cache = timeit(_perform_request, number=1)


def test_with_cache_faster() -> None:
    global time_without_cache, time_with_cache
    assert time_without_cache is not None
    assert time_with_cache is not None
    logger.debug(
        "Request took {:.4f}s without cache and {:.4f}s with cache.".format(
            time_without_cache, time_with_cache
        )
    )

    # Expect at least a 5x speedup through caching requests.
    assert time_without_cache > 5 * time_with_cache



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import traceback
from abc import abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Iterable, Mapping, Type, TypeVar, cast

from overrides import overrides
from typing_extensions import Final

from .consts import NASTY_DATE_TIME_FORMAT
from .io_ import read_file, read_lines_file, write_file, write_lines_file
from .typing_ import checked_cast

_T_JsonSerializable = TypeVar("_T_JsonSerializable", bound="JsonSerializable")


class JsonSerializable:
    @abstractmethod
    def to_json(self) -> Mapping[str, object]:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_json(
        cls: Type[_T_JsonSerializable], obj: Mapping[str, object]
    ) -> _T_JsonSerializable:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return type(self).__name__ + repr(self.to_json())


class JsonSerializedException(JsonSerializable):
    def __init__(
        self, time: datetime, type_: str, message: str, trace: Iterable[Iterable[str]]
    ):
        self.time: Final = time
        self.type: Final = type_
        self.message: Final = message
        self.trace: Final = trace

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    @overrides
    def to_json(self) -> Mapping[str, object]:
        return {
            "time": self.time.strftime(NASTY_DATE_TIME_FORMAT),
            "type": self.type,
            "message": self.message,
            "trace": self.trace,
        }

    @classmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "JsonSerializedException":
        return cls(
            time=datetime.strptime(
                checked_cast(str, obj["time"]), NASTY_DATE_TIME_FORMAT
            ),
            type_=checked_cast(str, obj["type"]),
            message=checked_cast(str, obj["message"]),
            trace=cast(Iterable[Iterable[str]], obj["trace"]),
        )

    @classmethod
    def from_exception(cls, exception: Exception) -> "JsonSerializedException":
        type_ = type(exception).__name__
        return cls(
            time=datetime.now(),
            type_=type_,
            message="{}: {}".format(type_, str(exception)),
            # rstrip/split() to be easier to read in formatted JSON.
            trace=[
                frame.rstrip().split("\n")
                for frame in traceback.format_tb(exception.__traceback__)
            ],
        )


def read_json(
    file: Path, type_: Type[_T_JsonSerializable], *, use_lzma: bool = False
) -> _T_JsonSerializable:
    return type_.from_json(json.loads(read_file(file, use_lzma=use_lzma)))


def write_json(
    file: Path,
    value: _T_JsonSerializable,
    *,
    overwrite_existing: bool = False,
    use_lzma: bool = False,
) -> None:
    write_file(
        file,
        json.dumps(value.to_json(), indent=2),
        overwrite_existing=overwrite_existing,
        use_lzma=use_lzma,
    )


def read_json_lines(
    file: Path, type_: Type[_T_JsonSerializable], *, use_lzma: bool = False
) -> Iterable[_T_JsonSerializable]:
    for line in read_lines_file(file, use_lzma=use_lzma):
        yield type_.from_json(json.loads(line))


def write_jsonl_lines(
    file: Path,
    values: Iterable[_T_JsonSerializable],
    *,
    overwrite_existing: bool = False,
    use_lzma: bool = False,
) -> None:
    write_lines_file(
        file,
        (json.dumps(value.to_json()) for value in values),
        overwrite_existing=overwrite_existing,
        use_lzma=use_lzma,
    )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from itertools import permutations
from pathlib import Path
from typing import Callable, Iterable, Mapping, Optional

import pytest
from _pytest.monkeypatch import MonkeyPatch

import nasty.batch.batch_results
from nasty._settings import NastySettings, TwitterApiSettings
from nasty.batch.batch import Batch
from nasty.batch.batch_results import BatchResults
from nasty.request.replies import Replies
from nasty.request.request import Request
from nasty.request.thread import Thread
from nasty.tweet.tweet import Tweet, TweetId


def _make_batch_results(
    settings: NastySettings,
    *,
    idify_dir: Optional[Path] = None,
    unidify_dir: Optional[Path] = None,
) -> BatchResults:
    batch = Batch()
    batch.append(Thread("1115689254271819777"))
    results = batch.execute()
    assert results is not None

    if idify_dir is not None:
        results = results.idify(idify_dir)
        assert results is not None

    if unidify_dir is not None:
        results = results.unidify(settings.twitter_api, unidify_dir)
        assert results is not None

    return results


def _assert_tweet_texts(results: BatchResults) -> None:
    assert [
        "We encourage submissions of new, previously, or concurrently published "
        "research. The event should be a forum for researchers to exchange ideas, "
        "discuss work, and get feedback. We hope you'll consider submitting your work.",
        "We'll have talks from research leaders on the latest advances in NLP. "
        "@NandoDF will be giving the keynote and more speakers will be announced soon. "
        "https://t.co/SB3URxn6ab",
        "Registration will open soon. In the meantime, we'll hope you'll save the date "
        "and consider joining us for what should be a fun day of listening to "
        "stimulating talks, mingling with like-minded people, exchanging ideas, and "
        "maybe even striking up a collaboration.",
    ] == [tweet.text for tweet in results.tweets(results[0])]


def _assert_tweet_ids(results: BatchResults) -> None:
    assert [
        TweetId("1115690002233556993"),
        TweetId("1115690615612825601"),
        TweetId("1115691710657499137"),
    ] == list(results.tweet_ids(results[0]))


def test_tweets_from_exectue(settings: NastySettings) -> None:
    results = _make_batch_results(settings)
    _assert_tweet_texts(results)
    _assert_tweet_ids(results)


def test_tweets_from_idify(settings: NastySettings, tmp_path: Path) -> None:
    results = _make_batch_results(settings, idify_dir=tmp_path)
    with pytest.raises(Exception):
        list(results.tweets(results[0]))
    _assert_tweet_ids(results)


def test_tweets_from_unidify(settings: NastySettings, tmp_path: Path) -> None:
    idify_dir = tmp_path / "idify"
    unidify_dir = tmp_path / "unidify"
    results = _make_batch_results(
        settings, idify_dir=idify_dir, unidify_dir=unidify_dir
    )
    _assert_tweet_texts(results)
    _assert_tweet_ids(results)


def test_samedir(settings: NastySettings, tmp_path: Path) -> None:
    idify_dir = tmp_path / "idify"
    unidify_dir = tmp_path / "unidify"
    results = _make_batch_results(settings, idify_dir=idify_dir)
    _assert_tweet_ids(results)
    results = _make_batch_results(
        settings, idify_dir=idify_dir, unidify_dir=unidify_dir
    )
    _assert_tweet_ids(results)
    _assert_tweet_texts(results)


def test_double_idify(settings: NastySettings, tmp_path: Path) -> None:
    results = _make_batch_results(settings, idify_dir=tmp_path).idify(tmp_path)
    assert results is not None
    with pytest.raises(Exception):
        list(results.tweets(results[0]))
    _assert_tweet_ids(results)


def test_double_unidify(settings: NastySettings, tmp_path: Path) -> None:
    idify_dir = tmp_path / "idify"
    unidify_dir = tmp_path / "unidify"

    results = _make_batch_results(
        settings, idify_dir=idify_dir, unidify_dir=unidify_dir
    )
    results = results.unidify(settings.twitter_api, unidify_dir)
    assert results is not None
    _assert_tweet_ids(results)
    _assert_tweet_texts(results)


def _mock_statuses_lookup(
    tweets: Mapping[TweetId, Tweet]
) -> Callable[[Iterable[TweetId], TwitterApiSettings], Iterable[Optional[Tweet]]]:
    def statuses_lookup(
        tweet_ids: Iterable[TweetId], twitter_api_settings: TwitterApiSettings
    ) -> Iterable[Tweet]:
        return (tweets[tweet_id] for tweet_id in tweet_ids)

    return statuses_lookup


@pytest.mark.parametrize(
    "requests",
    list(
        permutations(
            [
                Replies(TweetId("1115690002233556993")),
                Replies(TweetId("1115690615612825601")),
                Replies(TweetId("1115691710657499137")),
            ]
        )
    ),
    ids=repr,
)
def test_unidify_fail_and_restart(
    requests: Iterable[Request],
    settings: NastySettings,
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
) -> None:
    idify_dir = tmp_path / "idify"
    unidify_dir = tmp_path / "unidify"

    batch = Batch()
    for request in requests:
        batch.append(request)
    results = batch.execute()
    assert results is not None

    tweets = {tweet.id: tweet for entry in results for tweet in results.tweets(entry)}
    tweets_truncated = dict(tweets)
    del tweets_truncated[TweetId("1115690615612825601")]

    idified = results.idify(idify_dir)
    assert idified is not None

    monkeypatch.setattr(
        nasty.batch.batch_results,
        nasty.batch.batch_results.statuses_lookup.__name__,  # type: ignore
        _mock_statuses_lookup(tweets_truncated),
    )

    # Assert KeyError is propagated, because a Tweet is missing from tweets_truncated.
    with pytest.raises(KeyError):
        idified.unidify(settings.twitter_api, unidify_dir)
    unidified = BatchResults(unidify_dir)
    assert len(batch) > len(unidified)

    monkeypatch.setattr(
        nasty.batch.batch_results,
        nasty.batch.batch_results.statuses_lookup.__name__,  # type: ignore
        _mock_statuses_lookup(tweets),
    )

    unidified = idified.unidify(settings.twitter_api, unidify_dir)
    assert unidified is not None
    assert len(batch) == len(unidified)
    assert tweets == {
        tweet.id: tweet for entry in unidified for tweet in unidified.tweets(entry)
    }



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from overrides import overrides

from ..tweet.conversation_tweet_stream import ConversationTweetStream
from .conversation_request import ConversationRequest


class Replies(ConversationRequest):
    @overrides
    def request(self) -> ConversationTweetStream:
        from .._retriever.replies_retriever import RepliesRetriever

        return RepliesRetriever(self).tweet_stream



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Iterable, Optional, Sequence, Type, cast

from overrides import overrides
from typing_extensions import Final

from .._util.typing_ import checked_cast
from ..request.thread import Thread
from ..tweet.tweet import TweetId
from .conversation_retriever import ConversationRetriever, ConversationRetrieverBatch


class ThreadRetrieverBatch(ConversationRetrieverBatch):
    @overrides
    def _tweet_ids(self) -> Iterable[TweetId]:
        # TODO: ensure all tweets in a thread are by the same user

        # The first conversation batch contains entries for the Tweet with the requested
        # URL and possibly multiple conversation threads where the first on is always
        # the one where the Tweet author responds to himself for the first time:
        # {
        #     ...
        #     "timeline": {
        #         "id": "Conversation-1155486497451184128",
        #         "instructions": [
        #             {
        #                 "addEntries": {
        #                     "entries": [
        #                         {"entryId": "tweet-1155486497451184128", ...},
        #                         {"entryId": "conversationThread-11554883561...", ...},
        #                     ]
        #                 },
        #             },
        #             ...
        #         ],
        #         ...
        #     },
        # }
        #
        # Inside of the "conversationThread-" we find the actual thread Tweets:
        # {
        #     "entryId": "conversationThread-1155488356165398528",
        #     "sortIndex": "8067885539403591669",
        #     "content": {
        #         "timelineModule": {
        #             "items": [
        #                 {"entryId": "tweet-1155488356165398528", ...},
        #                 {"entryId": "tweet-1155490920621473792", ...},
        #                 ...
        #                 {"entryId": "conversationThread-11...-show_more_cursor", ...},
        #             ],
        #             ...
        #         }
        #     },
        # }
        # If the thread is over, the "...-show_more_cursor" is not present.
        #
        # All following batches will look like this instead:
        # {
        #     ...
        #     "timeline": {
        #         "id": "Conversation-1155486497451184128",
        #         "instructions": [
        #             {
        #                 "addToModule": {
        #                     "moduleItems": [
        #                         {"entryId": "tweet-1156606255940739078", ...},
        #                         {"entryId": "tweet-1156634771507810306", ...},
        #                         ...
        #                     ]
        #                 }
        #             }
        #         ],
        #     },
        # }
        for entry in self._parse_instructions():
            if entry["entryId"].startswith("tweet-"):
                # Tweets in the thread look like this:
                # {
                #     "entryId": "tweet-1155488356165398528",
                #     "item": {
                #         "content": {
                #             "tweet": {
                #                 "id": "1155488356165398528",
                #                 "displayType": "SelfThread",
                #             }
                #         },
                #         ...
                #     },
                # }

                if "tombstone" in entry["item"]["content"]:
                    # Sometimes Tweets become unavailable over time (for instance
                    # because they were deleted). They sometimes still show up
                    # in results but are only designated with a tombstone. We skip those
                    # when returning results.
                    # {
                    #     "entryId": "tweet-1079406406644715520",
                    #     "item": {
                    #         "content": {
                    #             "tombstone": {
                    #                 "displayType": "Inline",
                    #                 "tombstoneInfo": {
                    #                     "text": "",
                    #                     "richText": {
                    #                         "text": "This Tweet is unavailable.",
                    #                         "entities": [],
                    #                         "rtl": False,
                    #                     },
                    #                 },
                    #                 "epitaph": "Suspended",
                    #             }
                    #         },
                    #         ...
                    #     },
                    # }
                    self.num_tombstones += 1

                else:
                    yield entry["item"]["content"]["tweet"]["id"]

            elif entry["entryId"].startswith("conversationThread-") and entry[
                "entryId"
            ].endswith("-show_more_cursor"):
                pass
            elif entry["entryId"] == "novel_coronavirus_message":
                pass
            else:
                raise RuntimeError(
                    "Unknown entry type in entry-ID: '{}'.".format(entry["entryId"])
                )

    @overrides
    def _next_cursor(self) -> Optional[str]:
        # See the documentation of _tweet_ids_in_batch() on where the cursor entry
        # occurs. It looks like this:
        # {
        #     "entryId": "conversationThread-1155488...-show_more_cursor",
        #     "item": {
        #         "content": {
        #             "timelineCursor": {
        #                 "value": "TBwcFoLAvI3JhYuNIBUCAAAYJmNvbnZlcnNhdGlvbl...",
        #                 "cursorType": "ShowMore",
        #                 "displayTreatment": {"actionText": "5 more replies"},
        #             }
        #         },
        #         ...
        #     },
        # }
        instructions = self._parse_instructions()
        if not instructions:
            return None

        cursor_entry = instructions[-1]
        if cursor_entry["entryId"].startswith("conversationThread-") and cursor_entry[
            "entryId"
        ].endswith("-show_more_cursor"):
            return checked_cast(
                str, cursor_entry["item"]["content"]["timelineCursor"]["value"]
            )
        return None

    def _parse_instructions(self) -> Sequence[Any]:
        # See the documentation of _tweet_ids_in_batch() on what JSON-structures
        # this navigates.
        instructions: Final = cast(
            Sequence[Any], self._json["timeline"]["instructions"]
        )
        if "addEntries" in instructions[0]:
            entries = instructions[0]["addEntries"]["entries"]
            if not (
                len(entries) >= 2
                and entries[1]["entryId"].startswith("conversationThread-")
            ):
                return []
            return cast(Sequence[Any], entries[1]["content"]["timelineModule"]["items"])
        elif "addToModule" in instructions[0]:
            return cast(Sequence[Any], instructions[0]["addToModule"]["moduleItems"])
        else:
            raise RuntimeError("Could not parse conversation instructions.")


class ThreadRetriever(ConversationRetriever[Thread]):
    @classmethod
    @overrides
    def _retriever_batch_type(cls) -> Type[ThreadRetrieverBatch]:
        return ThreadRetrieverBatch



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional, Sequence, Tuple

import pytest

from nasty.request.thread import Thread
from nasty.tweet.tweet import TweetId


@pytest.mark.parametrize("max_tweets", [1, 10, 100], ids=repr)
def test_max_tweets(max_tweets: int) -> None:
    tweets = list(
        Thread(TweetId("1183715553057239040"), max_tweets=max_tweets).request()
    )
    assert max_tweets == len(tweets)
    assert len(tweets) == len({tweet.id for tweet in tweets})


@pytest.mark.parametrize("tweet_id", [TweetId("1110485791279595525")], ids=repr)
def test_no_thread(tweet_id: TweetId) -> None:
    assert not list(Thread(tweet_id).request())


@pytest.mark.parametrize(
    "args",
    [
        (
            TweetId("1115689254271819777"),
            [
                TweetId("1115690002233556993"),
                TweetId("1115690615612825601"),
                TweetId("1115691710657499137"),
            ],
        ),
    ],
    ids=repr,
)
def test_exact(args: Tuple[TweetId, Sequence[TweetId]]) -> None:
    tweet_id, thread = args
    assert thread == [tweet.id for tweet in Thread(tweet_id).request()]


@pytest.mark.parametrize(
    "args",
    [
        (TweetId("1155486497451184128"), 35, None),
        (TweetId("1180505950613958658"), 7, None),
    ],
    ids=repr,
)
def test_unlimited(args: Tuple[TweetId, int, Optional[int]]) -> None:
    tweet_id, min_expected, min_tombstones = args
    # Using batch_size=100 to speed up these larger requests.
    tweets = list(Thread(tweet_id, max_tweets=None, batch_size=100).request())
    assert min_expected <= len(tweets)
    # TODO: assert min_tombstones
    assert len(tweets) == len({tweet.id for tweet in tweets})



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from logging import getLogger
from typing import Any, Iterable, Mapping, Optional, Sequence, Type, cast

from overrides import overrides
from typing_extensions import Final

from .._util.typing_ import checked_cast
from ..request.search import Search, SearchFilter
from ..tweet.tweet import TweetId
from .retriever import Retriever, RetrieverBatch

logger = getLogger(__name__)


class SearchRetrieverBatch(RetrieverBatch):
    @overrides
    def _tweet_ids(self) -> Iterable[TweetId]:
        instructions: Final = cast(
            Sequence[Any], self._json["timeline"]["instructions"]
        )

        # Search results are contained in instructions. The first batch of a search will
        # look like this:
        # {
        #     ...
        #     "timeline": {
        #         "id": "search-6602913952152093875",
        #         "instructions": [
        #             {
        #                 "addEntries": {
        #                     "entries": [
        #                         {"entryId": "sq-I-t-1155486497451184128", ...},
        #                         {"entryId": "sq-I-t-1194473608061607936", ...},
        #                         {"entryId": "sq-M-1-d7721393", ...},
        #                         {"entryId": "sq-E-1981039365", ...},
        #                         ...
        #                         {"entryId": "sq-cursor-top", ...},
        #                         {"entryId": "sq-cursor-bottom", ...},
        #                     ]
        #                 }
        #             }
        #         ],
        #         ...
        #     },
        # }
        #
        # We need to separate the following entity types:
        # - "sq-I-t-..." are the Tweets matching the search query.
        # - "sq-M-..." contain supplementary information like user profiles that are
        #   somehow related to the matching Tweets (usually occurs once).
        # - "sq-E-..." seem to contain suggested live events (occur rarely).
        # - "sq-cursor-..." entries contain the cursors to fetch the next batch.
        #
        # All following batches will look similar except that the "sq-cursor-..."
        # entries are now differently placed:
        # {
        #     ...
        #     "timeline": {
        #         "id": "search-6602913956034868792",
        #         "instructions": [
        #             {
        #                 "addEntries": {
        #                     "entries": [
        #                         {"entryId": "sq-I-t-1157704001112219650", ...},
        #                         {"entryId": "sq-I-t-1156734175040266240", ...},
        #                         ...
        #                     ]
        #                 }
        #             },
        #             {
        #                 "replaceEntry": {
        #                     "entryIdToReplace": "sq-cursor-top",
        #                     "entry": {"entryId": "sq-cursor-top", ...},
        #                 }
        #             },
        #             {
        #                 "replaceEntry": {
        #                     "entryIdToReplace": "sq-cursor-bottom",
        #                     "entry": {"entryId": "sq-cursor-bottom", ...},
        #                 }
        #             },
        #         ],
        #         ...
        #     },
        # }

        for entry in instructions[0]["addEntries"]["entries"]:
            if entry["entryId"].startswith("sq-I-t-"):
                # Matching Tweet entries look like this:
                # {
                #     "entryId": "sq-I-t-1155486497451184128",
                #     "sortIndex": "999970",
                #     "content": {
                #         "item": {
                #             "content": {
                #                 "tweet": {
                #                     "id": "1155486497451184128",
                #                     "displayType": "Tweet",
                #                     "highlights": {...},
                #                 }
                #             },
                #             ...
                #         }
                #     },
                # }
                tweet = entry["content"]["item"]["content"]["tweet"]

                if "promotedMetadata" in tweet:
                    # Tweets which have been promoted look like this:
                    # {
                    #     "id": "1279090694313959431",
                    #     "displayType": "Tweet",
                    #     "promotedMetadata": {
                    #         "advertiserId": "177112193",
                    #         "impressionId": "14aeb0b7a6b63a42",
                    #         "disclosureType": "NoDisclosure",
                    #         "experimentValues": {},
                    #         "promotedTrendId": "0",
                    #     },
                    # }
                    logger.debug("Skipping promoted Tweet {}.".format(tweet["id"]))
                    continue

                yield checked_cast(TweetId, tweet["id"])
            elif entry["entryId"].startswith("sq-I-s-"):
                # These entries look like the following and seem to be related to
                # spell-checking the search queries, offering alternative suggestions,
                # and highlighting results, so we ignore them.
                # {
                #   "entryId": "sq-I-s-aca75bb6",
                #   "sortIndex": "999980",
                #   "content": {
                #     "item": {
                #       "content": {
                #         "spelling": {
                #           "spellingResult": {
                #             "text": "hilary",
                #             "hitHighlights": [
                #               {
                #                 "startIndex": 0,
                #                 "endIndex": 6
                #               }
                #             ],
                #             "score": 101.78357696533203
                #           },
                #           "spellingAction": "Expand",
                #           "originalQuery": "hillary lang:en"
                #         }
                #       },
                #       "clientEventInfo": {
                #         "component": "result",
                #         "element": "spelling"
                #       }
                #     }
                #   }
                # }
                pass
            elif entry["entryId"].startswith("sq-M-"):
                # These entries look like the following and seem to suggest users
                # related to the search term, so we ignore them.
                # {
                #   "entryId": "sq-M-1-a0d759d2",
                #   "sortIndex": "999970",
                #   "content": {
                #     "timelineModule": {
                #       "items": [
                #         {
                #           "entryId": "sq-MI-u-1339835893",
                #           "item": {
                #             "content": {
                #               "user": {
                #                 "id": "1339835893",
                #                 "displayType": "UserDetailed"
                #               }
                #             },
                #             "clientEventInfo": {
                #               "component": "user_module",
                #               "element": "user",
                #               "details": {
                #                 "timelinesDetails": {
                #                   "controllerData": "DAACDAAFDAABDAABDAACCgABAAAAAAAAAAAAAAwAAgoAAQAAAAAAAAABCgACACkc9QD8cSgLAAMAAAAPaGlsbGFyeSBsYW5nOmVuAAAAAAA="
                #                 }
                #               }
                #             }
                #           }
                #         },
                #         ...
                #       ],
                #       "displayType": "Vertical",
                #       "header": {
                #         "text": "People",
                #         "sticky": true
                #       },
                #       "footer": {
                #         "text": "View all",
                #         "url": "twitter://search?query=hillary+lang%3Aen&src=typed_query&type=users"
                #       },
                #       "clientEventInfo": {
                #         "component": "user_module",
                #         "element": "module"
                #       }
                #     }
                #   }
                # }
                pass
            elif entry["entryId"].startswith("sq-E-"):
                # TODO: document usage and example.
                pass
            elif entry["entryId"].startswith("sq-cursor-"):
                # These entries look like the following and are used to query the
                # previous/next Tweet batch. They are used in _next_cursor() but ignored
                # here.
                # {
                #   "entryId": "sq-cursor-top",
                #   "sortIndex": "999999999",
                #   "content": {
                #     "operation": {
                #       "cursor": {
                #         "value": "refresh:thGAVUV0VFVBYBFoCo98iIpPfgIxIYzAESY8LrAAAB9D-AYk3S8an8AAAAFxImKQvel2AHEiYpPuYXcAMSJiaHY1eQARE-MiL5FqABEiYlsPPXgAASJq0IQFTABBImicvYVgAQEiYSREbWoAESJkPX7xdwAhImPlMvl7AAEiYiUtlXsAMSIecAAVdgBhImRZcjF4ABEiZSHUTWkAQSJkUYiJeQBhIHM6qRVqAAEiZ34vSWsAASJipgjxewARImFK06V5AAEiaBkGQXcAESJifVa9agAhImkTQ-1gADEiYmF5rWkAElABUAJQARFaCFehWAiXoYBFVTRVIVABUAFS4VABUAAA==",
                #         "cursorType": "Top"
                #       }
                #     }
                #   }
                # },
                # {
                #   "entryId": "sq-cursor-bottom",
                #   "sortIndex": "0",
                #   "content": {
                #     "operation": {
                #       "cursor": {
                #         "value": "scroll:thGAVUV0VFVBYBFoCo98iIpPfgIxIYzAESY8LrAAAB9D-AYk3S8an8AAAAFxImKQvel2AHEiYpPuYXcAMSJiaHY1eQARE-MiL5FqABEiYlsPPXgAASJq0IQFTABBImicvYVgAQEiYSREbWoAESJkPX7xdwAhImPlMvl7AAEiYiUtlXsAMSIecAAVdgBhImRZcjF4ABEiZSHUTWkAQSJkUYiJeQBhIHM6qRVqAAEiZ34vSWsAASJipgjxewARImFK06V5AAEiaBkGQXcAESJifVa9agAhImkTQ-1gADEiYmF5rWkAElABUAJQARFaCFehWAiXoYBFVTRVIVABUAFS4VABUAAA==",
                #         "cursorType": "Bottom"
                #       }
                #     }
                #   }
                # }
                pass
            elif (
                entry["entryId"] == "novel_coronavirus_message"
                or entry["entryId"] == "novel_coronavirus_msg"
            ):
                # TODO: document usage and example.
                pass
            else:
                raise RuntimeError(
                    "Unknown entry type in entry-ID '{}'.".format(entry["entryId"])
                )

    @overrides
    def _next_cursor(self) -> Optional[str]:
        instructions: Final = cast(
            Sequence[Any], self._json["timeline"]["instructions"]
        )

        # As documented in _tweet_ids_in_batch(), the cursor objects can occur either
        # as part of "addEntries" or replaceEntry". We are only interested in
        # sq-cursor-bottom and I'm not sure what sq-cursor-top is for. The actual cursor
        # entry will look like this:
        # {
        #     "entryId": "sq-cursor-bottom",
        #     "sortIndex": "0",
        #     "content": {
        #         "operation": {
        #             "cursor": {"value": "scroll:thGAVUV...", "cursorType": "Bottom"}
        #         }
        #     },
        # }

        cursor_entry = None
        if instructions[0]["addEntries"]["entries"]:
            cursor_entry = instructions[0]["addEntries"]["entries"][-1]
        if not cursor_entry or cursor_entry["entryId"] != "sq-cursor-bottom":
            cursor_entry = instructions[-1]["replaceEntry"]["entry"]
        if cursor_entry["entryId"] != "sq-cursor-bottom":
            raise RuntimeError("Could not locate cursor entry.")
        return cast(
            Optional[str], cursor_entry["content"]["operation"]["cursor"]["value"]
        )


class SearchRetriever(Retriever[Search]):
    @classmethod
    @overrides
    def _retriever_batch_type(cls) -> Type[SearchRetrieverBatch]:
        return SearchRetrieverBatch

    @overrides
    def _timeline_url(self) -> Mapping[str, object]:
        return {
            "url": "https://mobile.twitter.com/search",
            "params": {
                "lang": self._request.lang,
                "q": self._q_url_param(),
                "src": "typed_query",
                "f": self._f_url_param(),
            },
        }

    @overrides
    def _batch_url(self) -> Mapping[str, object]:
        return {
            "url": "https://api.twitter.com/2/search/adaptive.json",
            "params": {
                # Not sure what most of the parameters with fixed values do. We set them
                # so that they are identical to those sent via the normal web interface.
                "include_profile_interstitial_type": "1",
                "include_blocking": "1",
                "include_blocked_by": "1",
                "include_followed_by": "1",
                "include_want_retweets": "1",
                "include_mute_edge": "1",
                "include_can_dm": "1",
                "include_can_media_tag": "1",
                "skip_status": "1",
                "cards_platform": "Web-12",
                "include_cards": "1",
                "include_composer_source": "true",
                "include_ext_alt_text": "true",
                "include_reply_count": "1",
                "tweet_mode": "extended",
                "include_entities": "true",
                "include_user_entities": "true",
                "include_ext_media_color": "true",
                "include_ext_media_availability": "true",
                "send_error_codes": "true",
                "q": self._q_url_param(),
                "count": self._request.batch_size,
                "result_filter": self._result_filter_url_param(),
                "tweet_search_mode": self._tweet_search_mode_url_param(),
                "query_source": "typed_query",
                "cursor": self._cursor,
                "pc": "1",
                "spelling_corrections": "1",
                "ext": "mediaStats,highlightedLabel,cameraMoment",
            },
        }

    def _q_url_param(self) -> str:
        """Transforms the stored query into the form that can be submitted to Twitter as
        an URL param.

        Does not perform URL escaping.
        """
        result = self._request.query
        if self._request.since:
            result += " since:" + self._request.since.isoformat()
        if self._request.until:
            result += " until:" + self._request.until.isoformat()
        result += " lang:" + self._request.lang
        return result

    def _f_url_param(self) -> Optional[str]:
        return {
            SearchFilter.LATEST: "live",
            SearchFilter.PHOTOS: "image",
            SearchFilter.VIDEOS: "video",
        }.get(self._request.filter, None)

    def _result_filter_url_param(self) -> Optional[str]:
        return {SearchFilter.PHOTOS: "image", SearchFilter.VIDEOS: "video"}.get(
            self._request.filter, None
        )

    def _tweet_search_mode_url_param(self) -> Optional[str]:
        return {SearchFilter.LATEST: "live"}.get(self._request.filter, None)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from logging import getLogger
from operator import itemgetter
from pathlib import Path
from typing import (
    Callable,
    Counter,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

from more_itertools import groupby_transform, spy, unzip

from nasty._settings import TwitterApiSettings

from .._util.io_ import read_lines_file, write_lines_file
from .._util.json_ import read_json, read_json_lines, write_json, write_jsonl_lines
from .._util.tweepy_ import statuses_lookup
from ..tweet.tweet import Tweet, TweetId
from ._execute_result import _ExecuteResult
from .batch_entry import BatchEntry

logger = getLogger(__name__)


class BatchResults(Sequence[BatchEntry]):
    def __init__(self, results_dir: Path):
        self._results_dir = results_dir
        self._entries: Sequence[BatchEntry] = [
            read_json(meta_file, BatchEntry)
            for meta_file in self._results_dir.iterdir()
            if meta_file.name.endswith(".meta.json")
        ]

    def tweets(self, entry: BatchEntry) -> Iterable[Tweet]:
        data_file = self._results_dir / entry.data_file_name
        ids_file = self._results_dir / entry.ids_file_name
        if not data_file.exists() and ids_file.exists():
            raise ValueError("Tweet data not available. Did you forget to unidify?")

        yield from read_json_lines(data_file, Tweet, use_lzma=True)

    def tweet_ids(self, entry: BatchEntry) -> Iterable[TweetId]:
        data_file = self._results_dir / entry.data_file_name
        ids_file = self._results_dir / entry.ids_file_name
        if ids_file.exists():
            yield from read_lines_file(ids_file)
        else:
            yield from (
                tweet.id for tweet in read_json_lines(data_file, Tweet, use_lzma=True)
            )

    def _transform(
        self,
        new_results_dir: Optional[Path],
        transform_name: str,
        transform_func: Callable[..., Counter[_ExecuteResult]],
        **transform_kwargs: object,
    ) -> Optional["BatchResults"]:
        results_dir = self._results_dir
        if new_results_dir is not None:
            results_dir = new_results_dir

        logger.debug(
            "{} batch results from '{}' to '{}'.".format(
                transform_name, self._results_dir, results_dir
            )
        )

        same_dir = results_dir.exists() and results_dir.samefile(self._results_dir)
        if not same_dir:
            Path.mkdir(results_dir, exist_ok=True, parents=True)

        result_counter = transform_func(results_dir, **transform_kwargs)
        logger.info(
            "  {} batch results completed. {:d} successful, {:d} skipped, {:d} "
            "failed.".format(
                transform_name,
                result_counter[_ExecuteResult.SUCCESS],
                result_counter[_ExecuteResult.SKIP],
                result_counter[_ExecuteResult.FAIL],
            )
        )

        if result_counter[_ExecuteResult.FAIL]:
            logger.error("  {} failed .".format(transform_name))
            return None
        if not same_dir:
            return BatchResults(results_dir)
        return self

    def idify(self, new_results_dir: Optional[Path] = None) -> Optional["BatchResults"]:
        return self._transform(new_results_dir, "Idifying", self._transform_idify)

    def _transform_idify(self, results_dir: Path) -> Counter[_ExecuteResult]:
        result_counter = Counter[_ExecuteResult]()
        for entry in self:
            try:
                ids_file = results_dir / entry.ids_file_name
                meta_file = results_dir / entry.meta_file_name

                if ids_file.exists() and meta_file.exists():
                    result_counter[_ExecuteResult.SKIP] += 1
                    continue

                write_lines_file(ids_file, self.tweet_ids(entry))
                write_json(meta_file, entry, overwrite_existing=True)
                result_counter[_ExecuteResult.SUCCESS] += 1
            except Exception:
                logger.exception("  Entry '{}' failed with exception.".format(entry.id))
                result_counter[_ExecuteResult.FAIL] += 1
        return result_counter

    def unidify(
        self,
        twitter_api_settings: TwitterApiSettings,
        new_results_dir: Optional[Path] = None,
    ) -> Optional["BatchResults"]:
        return self._transform(
            new_results_dir,
            "Unidifying",
            self._transform_unidify,
            twitter_api_settings=twitter_api_settings,
        )

    def _transform_unidify(
        self,
        results_dir: Path,
        twitter_api_settings: TwitterApiSettings,
    ) -> Counter[_ExecuteResult]:
        result_counter = Counter[_ExecuteResult]()

        head, entries_tweet_ids = spy(
            self._iter_entries_tweet_ids(results_dir, result_counter)
        )
        if not head:  # Check if any entries with Tweet-IDs exist (else unzip fails).
            return result_counter

        entries, tweet_ids = cast(
            Tuple[Iterator[BatchEntry], Iterator[TweetId]], unzip(entries_tweet_ids)
        )
        for entry, tweets in groupby_transform(
            zip(entries, statuses_lookup(tweet_ids, twitter_api_settings)),
            keyfunc=itemgetter(0),
            valuefunc=itemgetter(1),
        ):
            write_jsonl_lines(
                results_dir / entry.data_file_name,
                (tweet for tweet in tweets if tweet is not None),
                use_lzma=True,
            )
            write_json(
                results_dir / entry.meta_file_name, entry, overwrite_existing=True
            )
            result_counter[_ExecuteResult.SUCCESS] += 1

        return result_counter

    def _iter_entries_tweet_ids(
        self, results_dir: Path, result_counter: Counter[_ExecuteResult]
    ) -> Iterable[Tuple[BatchEntry, TweetId]]:
        for entry in self:
            meta_file = results_dir / entry.meta_file_name
            data_file = results_dir / entry.data_file_name
            if data_file.exists() and meta_file.exists():
                result_counter[_ExecuteResult.SKIP] += 1
                continue

            is_entry_empty = True
            for tweet in self.tweet_ids(entry):
                is_entry_empty = False
                yield entry, tweet

            if is_entry_empty:
                write_jsonl_lines(results_dir / entry.data_file_name, [], use_lzma=True)
                write_json(
                    results_dir / entry.meta_file_name, entry, overwrite_existing=True
                )
                result_counter[_ExecuteResult.SUCCESS] += 1

    def __len__(self) -> int:
        return len(self._entries)

    def __contains__(self, item: object) -> bool:
        return item in self._entries

    @overload
    def __getitem__(self, _index: int) -> BatchEntry:
        ...

    @overload  # noqa: F811
    def __getitem__(self, _slice: slice) -> Sequence[BatchEntry]:
        ...

    def __getitem__(  # noqa: F811
        self, index_or_slice: Union[int, slice]
    ) -> Union[BatchEntry, Sequence[BatchEntry]]:
        return self._entries[index_or_slice]

    def __iter__(self) -> Iterator[BatchEntry]:
        return iter(self._entries)

    def __repr__(self) -> str:
        return repr(self._entries)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path
from typing import Optional

from nasty._settings import TwitterApiSettings
from nasty.request.request import Request
from nasty.tweet.tweet import Tweet
from nasty.tweet.tweet_stream import TweetStream


class MockRequestContext:
    RESULT_TWEET = Tweet({})

    def __init__(self, *, num_results: int = 0):
        self.request: Optional[Request] = None
        self.remaining_result_tweets = num_results

        outer_self = self

        class MockTweetStream(TweetStream):
            def __next__(self) -> Tweet:
                if outer_self.remaining_result_tweets:
                    outer_self.remaining_result_tweets -= 1
                    return outer_self.RESULT_TWEET
                raise StopIteration()

        def mock_request(request: Request) -> TweetStream:
            self.request = request
            if request.max_tweets:
                self.remaining_result_tweets = min(
                    self.remaining_result_tweets, request.max_tweets
                )
            return MockTweetStream()

        self.mock_request = mock_request


class MockBatchContext:
    def __init__(self) -> None:
        self.load_args = None
        self.execute_args = None

        class MockBatch:
            @staticmethod
            def load(file: Path) -> None:
                self.load_args = (file,)

            @staticmethod
            def execute(results_dir: Optional[Path]) -> None:
                self.execute_args = (results_dir,)

        self.MockBatch = MockBatch


class MockBatchResultsContext:
    def __init__(self) -> None:
        self.init_args = None
        self.idify_args = None
        self.unidify_args = None

        class MockBatchResults:
            @staticmethod
            def __init__(results_dir: Path):
                self.init_args = (results_dir,)

            @staticmethod
            def idify(results_dir: Path) -> None:
                self.idify_args = (results_dir,)

            @staticmethod
            def unidify(
                twitter_api_settings: TwitterApiSettings,
                results_dir: Path,
            ) -> None:
                self.unidify_args = (results_dir,)

        self.MockBatchResults = MockBatchResults



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import sys
from datetime import date
from pathlib import Path
from typing import Mapping, Optional

from nasty_utils import (
    Argument,
    ArgumentGroup,
    Program,
    ProgramConfig,
    checked_cast,
    parse_yyyy_mm_dd,
)
from overrides import overrides
from pydantic import validator

import nasty
from nasty._settings import NastySettings
from nasty._util.tweepy_ import statuses_lookup
from nasty.batch.batch import Batch
from nasty.batch.batch_results import BatchResults
from nasty.request.replies import Replies
from nasty.request.request import DEFAULT_BATCH_SIZE, Request
from nasty.request.search import DEFAULT_FILTER, Search, SearchFilter
from nasty.request.thread import Thread
from nasty.tweet.tweet import Tweet, TweetId

# TODO: Order Argument Groups


_REQUEST_ARGUMENT_GROUP = ArgumentGroup(
    name="Request Arguments",
    description="Control how Tweets are requested.",
)
_BATCH_ARGUMENT_GROUP = ArgumentGroup(
    name="Batch Arguments",
    description=(
        "NASTY supports storing requests in a batch file and executing them later via "
        "the 'nasty batch' command, which allows to operate in batch mode, track "
        "progress, and rerun uncompleted/failed requests."
    ),
)


class RequestProgram(Program):
    max_tweets: Optional[int] = Argument(
        100,
        alias="max-tweets",
        short_alias="n",
        description=(
            "Maximum number of tweets to retrieve. Set to -1 to receive as many as "
            "possible. Defaults to 100."
        ),
        metavar="N",
        group=_REQUEST_ARGUMENT_GROUP,
    )

    @validator("max_tweets")
    def _max_tweets_validator(cls, v: Optional[int]) -> Optional[int]:  # noqa: N805
        return v if v != -1 else None

    batch_size: int = Argument(
        -1,
        alias="batch-size",
        short_alias="i",
        description=(
            "Batch size to retrieve Tweets in. Set to -1 for default behavior. Only "
            "change when necessary."
        ),
        metavar="N",
        group=_REQUEST_ARGUMENT_GROUP,
    )

    @validator("batch_size")
    def _batch_size_validator(cls, v: int) -> int:  # noqa: N805
        return v if v != -1 else DEFAULT_BATCH_SIZE

    to_batch: Optional[Path] = Argument(
        alias="to-batch",
        short_alias="b",
        description="Append request to batch file instead of executing.",
        metavar="FILE",
        group=_BATCH_ARGUMENT_GROUP,
    )

    @overrides
    def run(self) -> None:
        request = self._build_request()
        if self.to_batch:
            batch = Batch()
            if self.to_batch.exists():
                batch.load(self.to_batch)
            self._batch_submit(batch, request)
            batch.dump(self.to_batch)
        else:
            for tweet in request.request():
                sys.stdout.write(json.dumps(tweet.to_json()) + "\n")

    def _build_request(self) -> Request:
        raise NotImplementedError()

    def _batch_submit(self, batch: Batch, request: Request) -> None:
        batch.append(request)


_SEARCH_ARGUMENT_GROUP = ArgumentGroup(
    name="Search Arguments", description="Control what kind of Tweets are searched."
)


class SearchProgram(RequestProgram):
    class Config(ProgramConfig):
        title = "search"
        aliases = ("s",)
        description = "Retrieve Tweets using the Twitter advanced search."

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    query: str = Argument(
        short_alias="q",
        description="Search string (required).",
        group=_SEARCH_ARGUMENT_GROUP,
    )

    since: Optional[date] = Argument(
        short_alias="s",
        description="Earliest date for Tweets (inclusive) as YYYY-MM-DD.",
        metavar="DATE",
        group=_SEARCH_ARGUMENT_GROUP,
    )

    @validator("since", pre=True)
    def _since_validator(cls, v: Optional[str]) -> Optional[date]:  # noqa: N805
        return parse_yyyy_mm_dd(v) if v else None

    until: Optional[date] = Argument(
        short_alias="u",
        description="Latest date for Tweets (exclusive) as YYYY-MM-DD.",
        metavar="DATE",
        group=_SEARCH_ARGUMENT_GROUP,
    )

    @validator("until", pre=True, always=False)
    def _until_validator(cls, v: Optional[str]) -> Optional[date]:  # noqa: N805
        return parse_yyyy_mm_dd(v) if v else None

    filter_: SearchFilter = Argument(
        DEFAULT_FILTER,
        alias="filter",
        short_alias="f",
        description=(
            "Sorting/filtering of Tweets (top, latest, photos, videos). Defaults "
            "to 'top'."
        ),
        group=_SEARCH_ARGUMENT_GROUP,
    )

    lang: str = Argument(
        "en",
        short_alias="l",
        description=(
            "Language for Tweets, presumably as ISO 3166-1 two or three letter codes. "
            "Defaults to 'en'."
        ),
        group=_SEARCH_ARGUMENT_GROUP,
    )

    daily: bool = Argument(
        False,
        short_alias="d",
        description=(
            "For a request with since and until date, append one search request "
            "per day in the date-range with identical settings otherwise."
        ),
        group=_BATCH_ARGUMENT_GROUP,
    )

    @validator("daily")
    def _daily_validator(
        cls, v: bool, values: Mapping[str, object]  # noqa:N805
    ) -> bool:
        if v and not values["to_batch"]:
            raise ValueError("-d/--daily requires -b/--to-batch.")
        if v and (values["since"] is None or values["until"] is None):
            raise ValueError("-d/--daily requires -s/--since and -u/--until.")
        return v

    @overrides
    def _build_request(self) -> Search:
        return Search(
            self.query,
            since=self.since,
            until=self.until,
            filter_=self.filter_,
            lang=self.lang,
            max_tweets=self.max_tweets,
            batch_size=self.batch_size,
        )

    @overrides
    def _batch_submit(self, batch: Batch, request: Request) -> None:
        request = checked_cast(Search, request)
        if self.daily:
            for daily_request in request.to_daily_requests():
                super()._batch_submit(batch, daily_request)
        else:
            super()._batch_submit(batch, request)


_REPLIES_ARGUMENT_GROUP = ArgumentGroup(
    name="Replies Arguments",
    description="Control to which Tweet replies are retrieved.",
)


class RepliesProgram(RequestProgram):
    class Config(ProgramConfig):
        title = "replies"
        aliases = ("r",)
        description = "Retrieve all directly replying Tweets to a Tweet."

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    tweet_id: TweetId = Argument(
        alias="tweet-id",
        short_alias="t",
        description="ID of the Tweet to retrieve replies for (required).",
        metavar="ID",
        group=_REPLIES_ARGUMENT_GROUP,
    )

    @overrides
    def _build_request(self) -> Request:
        return Replies(
            self.tweet_id, max_tweets=self.max_tweets, batch_size=self.batch_size
        )


_THREAD_ARGUMENT_GROUP = ArgumentGroup(
    name="Thread Arguments",
    description="Control to which Tweet's threaded Tweets are retrieved.",
)


class ThreadProgram(RequestProgram):
    class Config(ProgramConfig):
        title = "thread"
        aliases = ("t",)
        description = "Retrieve all Tweets threaded under a Tweet."

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    tweet_id: TweetId = Argument(
        alias="tweet-id",
        short_alias="t",
        description="ID of the Tweet to retrieve threaded Tweets for (required).",
        metavar="ID",
        group=_THREAD_ARGUMENT_GROUP,
    )

    @overrides
    def _build_request(self) -> Request:
        return Thread(
            self.tweet_id, max_tweets=self.max_tweets, batch_size=self.batch_size
        )


_BATCH_ARGUMENT_GROUP = ArgumentGroup(
    name="Batch Arguments",
    description="Execute previously created batch of requests.",
)


class BatchProgram(Program):
    class Config(ProgramConfig):
        title = "batch"
        aliases = ("b",)
        description = "Execute previously created batch of requests."

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    batch_file: Path = Argument(
        alias="batch-file",
        short_alias="b",
        description="Batch file to which requests have been appended.",
        metavar="FILE",
        group=_BATCH_ARGUMENT_GROUP,
    )

    results_dir: Path = Argument(
        alias="results-dir",
        short_alias="r",
        description="Directory to which results will be written.",
        metavar="DIR",
        group=_BATCH_ARGUMENT_GROUP,
    )

    @overrides
    def run(self) -> None:
        batch = Batch()
        batch.load(self.batch_file)
        batch.execute(self.results_dir)


_IDIFY_ARGUMENT_GROUP = ArgumentGroup(
    name="Idifiy Arguments",
    description=(
        "By default Tweets are read line-by-line from stdin and the corresponding "
        "Tweet-IDs are written to stdout. Use the following arguments to instead "
        "operate on a result directory of a batch of requests."
    ),
)


class IdifyProgram(Program):
    class Config(ProgramConfig):
        title = "idify"
        aliases = ("i", "id")
        description = "Reduce Tweet-collection to Tweet-IDs (for publishing)."

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    in_dir: Optional[Path] = Argument(
        alias="in-dir",
        short_alias="i",
        description="Directory with results of a batch of requests.",
        metavar="DIR",
        group=_IDIFY_ARGUMENT_GROUP,
    )

    out_dir: Optional[Path] = Argument(
        alias="out-dir",
        short_alias="o",
        description=(
            "Directory to which Tweet-IDs will be written. If not given, will use "
            "input directory."
        ),
        metavar="DIR",
        group=_IDIFY_ARGUMENT_GROUP,
    )

    @validator("out_dir")
    def _out_dir_validator(
        cls, v: Optional[Path], values: Mapping[str, object]  # noqa: N805
    ) -> Optional[Path]:
        if v and not values["in_dir"]:
            raise ValueError("-o/--out-dir requires -i/--in-dir.")
        return v

    @overrides
    def run(self) -> None:
        if self.in_dir:
            batch_results = BatchResults(self.in_dir)
            batch_results.idify(self.out_dir if self.out_dir else self.in_dir)
        else:
            for line in sys.stdin:
                sys.stdout.write(str(Tweet(json.loads(line)).id) + "\n")


_UNIDIFY_ARGUMENT_GROUP = ArgumentGroup(
    name="Unidify Arguments",
    description=(
        "By default Tweet-IDs are read line-by-line from stdin and collected Tweets "
        "are written to stdout. Use the following arguments to instead operate on "
        "idified batch results."
    ),
)


class UnidifyProgram(Program):
    class Config(ProgramConfig):
        title = "unidify"
        aliases = ("u", "unid")
        description = (
            "Collect full Tweet information from Tweet-IDs (via official Twitter API)."
        )

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )

    in_dir: Optional[Path] = Argument(
        alias="in-dir",
        short_alias="i",
        description="Directory with idified batch results.",
        metavar="DIR",
        group=_UNIDIFY_ARGUMENT_GROUP,
    )

    out_dir: Optional[Path] = Argument(
        alias="out-dir",
        short_alias="o",
        description=(
            "Directory to which unidified batch results will be written. If not "
            "given, will use input directory."
        ),
        metavar="DIR",
        group=_UNIDIFY_ARGUMENT_GROUP,
    )

    @validator("out_dir")
    def _out_dir_validator(
        cls, v: Optional[Path], values: Mapping[str, object]  # noqa: N805
    ) -> Optional[Path]:
        if v and not values["in_dir"]:
            raise ValueError("-o/--out-dir requires -i/--in-dir.")
        return v

    @overrides
    def run(self) -> None:
        if self.in_dir:
            batch_results = BatchResults(self.in_dir)
            batch_results.unidify(
                self.settings.twitter_api,
                self.out_dir if self.out_dir else self.in_dir,
            )
        else:
            for tweet in statuses_lookup(
                (TweetId(line.strip()) for line in sys.stdin), self.settings.twitter_api
            ):
                if tweet is not None:
                    sys.stdout.write(json.dumps(tweet.to_json()) + "\n")


class NastyProgram(Program):
    class Config(ProgramConfig):
        title = "nasty"
        version = nasty.__version__
        description = "NASTY Advanced Search Tweet Yielder."
        subprograms = (
            SearchProgram,
            RepliesProgram,
            ThreadProgram,
            BatchProgram,
            IdifyProgram,
            UnidifyProgram,
        )

    settings: NastySettings = Argument(
        alias="config", description="Overwrite default config file path."
    )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import datetime
from pathlib import Path
from typing import Mapping, Optional, cast

from overrides import overrides
from typing_extensions import Final

from .._util.consts import NASTY_DATE_TIME_FORMAT
from .._util.json_ import JsonSerializable, JsonSerializedException
from .._util.typing_ import checked_cast
from ..request.request import Request

BatchEntryId = str


class BatchEntry(JsonSerializable):
    def __init__(
        self,
        request: Request,
        *,
        id_: BatchEntryId,
        completed_at: Optional[datetime],
        exception: Optional[JsonSerializedException],
    ):
        self.request: Final = request
        self.id: Final = id_
        self.completed_at = completed_at
        self.exception = exception

    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    @property
    def meta_file_name(self) -> Path:
        return Path("{:s}.meta.json".format(self.id))

    @property
    def data_file_name(self) -> Path:
        return Path("{:s}.data.jsonl.xz".format(self.id))

    @property
    def ids_file_name(self) -> Path:
        return Path("{:s}.ids".format(self.id))

    @overrides
    def to_json(self) -> Mapping[str, object]:
        obj = {
            "id": self.id,
            "request": self.request.to_json(),
        }
        if self.completed_at:
            obj["completed_at"] = self.completed_at.strftime(NASTY_DATE_TIME_FORMAT)
        if self.exception is not None:
            obj["exception"] = self.exception.to_json()
        return obj

    @classmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "BatchEntry":
        return cls(
            request=Request.from_json(cast(Mapping[str, object], obj["request"])),
            id_=checked_cast(str, obj["id"]),
            completed_at=(
                datetime.strptime(
                    checked_cast(str, obj["completed_at"]), NASTY_DATE_TIME_FORMAT
                )
                if "completed_at" in obj
                else None
            ),
            exception=(
                JsonSerializedException.from_json(
                    cast(Mapping[str, object], obj["exception"])
                )
                if "exception" in obj
                else None
            ),
        )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABC, abstractmethod
from typing import Dict, Mapping, Optional

from overrides import overrides
from typing_extensions import Final, final

from .._util.json_ import JsonSerializable
from ..tweet.tweet_stream import TweetStream

DEFAULT_MAX_TWEETS: Final = 100
DEFAULT_BATCH_SIZE: Final = 20


class Request(ABC, JsonSerializable):
    def __init__(self, *, max_tweets: Optional[int], batch_size: int):
        """Construct a new timeline view.

        :param max_tweets: Stop retrieving Tweets after this many tweets have
            been found. Set to None in order to receive as many Tweets as
            possible. Note that this can return quite a lot of tweets,
            especially if using Search, Filter.LATEST and no date range.
        :param batch_size: The batch size in which Tweets should be retrieved.

            The normal web interface always queries 20 Tweets per batch. Twitter
            interprets this parameter more as a guideline and can either return
            more or less then the requested amount. This does not indicate that
            no more matching Tweets exist after this batch.

            Note that by setting anything unequal to 20 here, we make ourselves
            easily distinguishable from a normal web browser. Additionally,
            advanced queries like using AND or OR seem to no longer work as
            intended. For Thread and Reply, increasing the batch_size is likely
            to also increase the number of results (no idea why Twitter is doing
            this).

            This parameter can be used to speed up the retrieval performance, by
            reducing the HTTP overhead as less requests have to be performed per
            returned Tweet. If you want to do this, we identified 100 to be a
            good value because increasing it further does seem not return more
            Tweets per request.
        """

        if max_tweets is not None and max_tweets <= 0:
            raise ValueError("If max_tweets is given, it must be positive.")
        if batch_size <= 0:
            raise ValueError("batch_size must be positive.")

        self.max_tweets: Final = max_tweets
        self.batch_size: Final = batch_size

    @final
    @overrides
    def __eq__(self, other: object) -> bool:
        return type(self) == type(other) and self.__dict__ == other.__dict__

    @abstractmethod
    @overrides
    def to_json(self) -> Mapping[str, object]:
        obj: Dict[str, object] = {
            "type": type(self).__name__,
        }
        if self.max_tweets != DEFAULT_MAX_TWEETS:
            obj["max_tweets"] = self.max_tweets
        if self.batch_size != DEFAULT_BATCH_SIZE:
            obj["batch_size"] = self.batch_size
        return obj

    @classmethod
    @abstractmethod
    @overrides
    def from_json(cls, obj: Mapping[str, object]) -> "Request":
        from .replies import Replies
        from .search import Search
        from .thread import Thread

        if obj["type"] == Search.__name__:
            return Search.from_json(obj)
        elif obj["type"] == Replies.__name__:
            return Replies.from_json(obj)
        elif obj["type"] == Thread.__name__:
            return Thread.from_json(obj)

        raise RuntimeError("Unknown request type: '{}'.".format(obj["type"]))

    @abstractmethod
    def request(self) -> TweetStream:
        raise NotImplementedError()



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path

import pytest
from _pytest.capture import CaptureFixture
from _pytest.monkeypatch import MonkeyPatch

import nasty._cli
from nasty import main

from .mock_context import MockBatchContext


def test_correct_call(
    monkeypatch: MonkeyPatch,
    capsys: CaptureFixture,
    tmp_path: Path,
) -> None:
    mock_context = MockBatchContext()
    monkeypatch.setattr(
        nasty._cli,
        nasty._cli.Batch.__name__,  # type: ignore
        mock_context.MockBatch,
    )

    batch_file = tmp_path / "batch.jsonl"
    results_dir = tmp_path / "out"
    main("batch", "--batch-file", str(batch_file), "--results-dir", str(results_dir))

    assert mock_context.load_args == (batch_file,)
    assert mock_context.execute_args == (results_dir,)
    assert capsys.readouterr().out == ""


def test_no_batch_file(tmp_path: Path) -> None:
    batch_file = tmp_path / "batch.jsonl"
    results_dir = tmp_path / "out"

    with pytest.raises(FileNotFoundError):
        main(
            "batch",
            "--batch-file",
            str(batch_file),
            "--results-dir",
            str(results_dir),
        )



_.error_on_external_run  # unused attribute (noxfile.py:20)
_.reuse_existing_virtualenvs  # unused attribute (noxfile.py:21)
_.stop_on_first_error  # unused attribute (noxfile.py:22)
test  # unused function (noxfile.py:25)
_max_tweets_validator  # unused function (src/nasty/_cli.py:74)
_batch_size_validator  # unused function (src/nasty/_cli.py:90)
title  # unused variable (src/nasty/_cli.py:129)
aliases  # unused variable (src/nasty/_cli.py:130)
description  # unused variable (src/nasty/_cli.py:131)
_since_validator  # unused function (src/nasty/_cli.py:150)
_until_validator  # unused function (src/nasty/_cli.py:161)
_daily_validator  # unused function (src/nasty/_cli.py:196)
title  # unused variable (src/nasty/_cli.py:236)
aliases  # unused variable (src/nasty/_cli.py:237)
description  # unused variable (src/nasty/_cli.py:238)
title  # unused variable (src/nasty/_cli.py:267)
aliases  # unused variable (src/nasty/_cli.py:268)
description  # unused variable (src/nasty/_cli.py:269)
title  # unused variable (src/nasty/_cli.py:297)
aliases  # unused variable (src/nasty/_cli.py:298)
description  # unused variable (src/nasty/_cli.py:299)
title  # unused variable (src/nasty/_cli.py:340)
aliases  # unused variable (src/nasty/_cli.py:341)
description  # unused variable (src/nasty/_cli.py:342)
_out_dir_validator  # unused function (src/nasty/_cli.py:367)
title  # unused variable (src/nasty/_cli.py:397)
aliases  # unused variable (src/nasty/_cli.py:398)
description  # unused variable (src/nasty/_cli.py:399)
_out_dir_validator  # unused function (src/nasty/_cli.py:426)
title  # unused variable (src/nasty/_cli.py:452)
version  # unused variable (src/nasty/_cli.py:453)
description  # unused variable (src/nasty/_cli.py:454)
subprograms  # unused variable (src/nasty/_cli.py:455)
_.num_tombstones  # unused attribute (src/nasty/_retriever/conversation_retriever.py:36)
_.num_tombstones  # unused attribute (src/nasty/_retriever/replies_retriever.py:139)
_.num_tombstones  # unused attribute (src/nasty/_retriever/thread_retriever.py:132)
SingleMetavarHelpFormatter  # unused class (src/nasty/_util/argparse_.py:23)
_._format_action_invocation  # unused method (src/nasty/_util/argparse_.py:24)
pytest_configure  # unused function (tests/conftest.py:30)
activate_requests_cache  # unused function (tests/conftest.py:56)
disrespect_robotstxt  # unused function (tests/conftest.py:67)
min_tombstones  # unused variable (tests/retriever/test_replies.py:70)
min_tombstones  # unused variable (tests/retriever/test_thread.py:67)
exc_type  # unused variable (tests/util/requests_cache.py:174)
exc_val  # unused variable (tests/util/requests_cache.py:175)
exc_tb  # unused variable (tests/util/requests_cache.py:176)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from datetime import date
from logging import getLogger
from pathlib import Path
from typing import List, Mapping, Optional, Sequence, Type

import pytest
from _pytest.capture import CaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from typing_extensions import Final

from nasty import main
from nasty.batch.batch import Batch
from nasty.request.replies import Replies
from nasty.request.request import DEFAULT_BATCH_SIZE, DEFAULT_MAX_TWEETS, Request
from nasty.request.search import DEFAULT_FILTER, DEFAULT_LANG, Search, SearchFilter
from nasty.request.thread import Thread

from .mock_context import MockRequestContext

logger = getLogger(__name__)

REQUESTS: Final[Mapping[Type[Request], Sequence[Request]]] = {
    Search: [
        Search("trump"),
        Search("donald trump"),
        Search("trump", since=date(2019, 3, 21), until=date(2019, 3, 22)),
        Search("trump", filter_=SearchFilter.LATEST),
        Search("trump", lang="de"),
        Search("trump", max_tweets=17, batch_size=71),
        Search("trump", max_tweets=None, batch_size=DEFAULT_BATCH_SIZE),
    ],
    Replies: [
        Replies("332308211321425920"),
        Replies("332308211321425920", max_tweets=17, batch_size=71),
        Replies("332308211321425920", max_tweets=None, batch_size=DEFAULT_BATCH_SIZE),
    ],
    Thread: [
        Thread("332308211321425920"),
        Thread("332308211321425920", max_tweets=17, batch_size=71),
        Thread("332308211321425920", max_tweets=None, batch_size=DEFAULT_BATCH_SIZE),
    ],
}
ALL_REQUESTS: Final[Sequence[Request]] = [
    request for requests_for_type in REQUESTS.values() for request in requests_for_type
]


def _make_args(  # noqa: C901
    request: Request,
    to_batch: Optional[Path] = None,
    daily: bool = False,
) -> Sequence[str]:
    args: List[str] = []

    if isinstance(request, Search):
        args += ["search", "--query", request.query]
        if request.since:
            args += ["--since", request.since.strftime("%Y-%m-%d")]
        if request.until:
            args += ["--until", request.until.strftime("%Y-%m-%d")]
        if request.filter != DEFAULT_FILTER:
            args += ["--filter", request.filter.name]
        if request.lang != DEFAULT_LANG:
            args += ["--lang", request.lang]

    elif isinstance(request, Replies):
        args += ["replies", "--tweet-id", request.tweet_id]

    elif isinstance(request, Thread):
        args += ["thread", "--tweet-id", request.tweet_id]

    else:
        raise ValueError("Unimplemented Request subclass: {}.".format(type(request)))

    if request.max_tweets is None:
        args += ["--max-tweets", "-1"]
    elif request.max_tweets != DEFAULT_MAX_TWEETS:
        args += ["--max-tweets", str(request.max_tweets)]

    if request.batch_size != DEFAULT_BATCH_SIZE:
        args += ["--batch-size", str(request.batch_size)]

    if to_batch is not None:
        args += ["--to-batch", str(to_batch)]

    if daily:
        if not isinstance(request, Search):
            raise ValueError("daily can only be used for Search-requests.")
        args += ["--daily"]

    return args


@pytest.mark.parametrize("request_", ALL_REQUESTS, ids=repr)
def test_correct_call(
    request_: Request, monkeypatch: MonkeyPatch, capsys: CaptureFixture
) -> None:
    mock_context: MockRequestContext = MockRequestContext()
    monkeypatch.setattr(
        type(request_), request_.request.__name__, mock_context.mock_request
    )

    main(*_make_args(request_))

    assert mock_context.request == request_
    assert not mock_context.remaining_result_tweets
    assert capsys.readouterr().out == ""


@pytest.mark.parametrize("num_results", [5, 10, 20], ids=repr)
@pytest.mark.parametrize(
    "request_",
    [
        Search("trump", max_tweets=10),
        Replies("332308211321425920", max_tweets=10),
        Thread("332308211321425920", max_tweets=10),
    ],
    ids=repr,
)
def test_correct_call_results(
    request_: Request,
    num_results: int,
    monkeypatch: MonkeyPatch,
    capsys: CaptureFixture,
) -> None:
    mock_context: MockRequestContext = MockRequestContext(num_results=num_results)
    monkeypatch.setattr(
        type(request_), request_.request.__name__, mock_context.mock_request
    )

    main(*_make_args(request_))

    assert mock_context.request == request_
    assert not mock_context.remaining_result_tweets
    assert capsys.readouterr().out == (
        json.dumps(mock_context.RESULT_TWEET.to_json()) + "\n"
    ) * min(10, num_results)


@pytest.mark.parametrize("request_", ALL_REQUESTS, ids=repr)
def test_correct_call_to_batch(
    request_: Request,
    capsys: CaptureFixture,
    tmp_path: Path,
) -> None:
    batch_file = tmp_path / "batch.jsonl"

    main(*_make_args(request_, to_batch=batch_file))

    assert capsys.readouterr().out == ""
    batch = Batch()
    batch.load(batch_file)
    assert len(batch) == 1
    assert batch[0].request == request_
    assert batch[0].id
    assert batch[0].completed_at is None
    assert batch[0].exception is None


@pytest.mark.parametrize(
    "old_request,new_request",
    [
        (REQUESTS[Search][0], REQUESTS[Search][1]),
        (REQUESTS[Replies][0], REQUESTS[Replies][1]),
        (REQUESTS[Thread][0], REQUESTS[Thread][1]),
    ],
    ids=repr,
)
def test_correct_call_to_batch_exists(
    old_request: Request,
    new_request: Request,
    capsys: CaptureFixture,
    tmp_path: Path,
) -> None:
    batch_file = tmp_path / "batch.jsonl"
    batch = Batch()
    batch.append(old_request)
    batch.dump(batch_file)

    main(*_make_args(new_request, to_batch=batch_file))

    assert capsys.readouterr().out == ""
    batch = Batch()
    batch.load(batch_file)
    assert len(batch) == 2
    for batch_entry, expected_request in zip(batch, [old_request, new_request]):
        assert batch_entry.request == expected_request
        assert batch_entry.id
        assert batch_entry.completed_at is None
        assert batch_entry.exception is None


def test_correct_call_to_batch_daily(capsys: CaptureFixture, tmp_path: Path) -> None:
    batch_file = tmp_path / "batch.jsonl"
    request = Search("trump", since=date(2019, 1, 1), until=date(2019, 2, 1))

    # Needed for type checking.
    assert request.until is not None and request.since is not None

    main(*_make_args(request, to_batch=batch_file, daily=True))

    assert capsys.readouterr().out == ""
    batch = Batch()
    batch.load(batch_file)
    assert len(batch) == (request.until - request.since).days
    for batch_entry, expected_request in zip(batch, request.to_daily_requests()):
        assert batch_entry.request == expected_request
        assert batch_entry.id
        assert batch_entry.completed_at is None
        assert batch_entry.exception is None



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import sys
from copy import deepcopy
from io import StringIO
from pathlib import Path

from _pytest.capture import CaptureFixture
from _pytest.monkeypatch import MonkeyPatch

import nasty._cli
from nasty import main
from nasty.tweet.tweet import Tweet

from ..test_tweet import tweet_jsons
from .mock_context import MockBatchResultsContext


def test_idify_stdin(monkeypatch: MonkeyPatch, capsys: CaptureFixture) -> None:
    tweet_json = tweet_jsons["1142944425502543875"]
    tweets = []
    for i in range(5):
        tweet = deepcopy(tweet_json)
        tweet["id"] = i
        tweet["id_str"] = str(i)
        tweets.append(tweet)

    monkeypatch.setattr(
        sys,
        "stdin",
        StringIO("\n".join(json.dumps(tweet) for tweet in tweets)),
    )
    main("idify")
    assert capsys.readouterr().out == "\n".join(str(i) for i in range(5)) + "\n"


def test_idify_indir(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    mock_context = MockBatchResultsContext()
    monkeypatch.setattr(
        nasty._cli,
        nasty._cli.BatchResults.__name__,  # type: ignore
        mock_context.MockBatchResults,
    )

    main("idify", "--in-dir", str(tmp_path))

    assert mock_context.init_args == (tmp_path,)
    assert mock_context.idify_args == (tmp_path,)
    assert mock_context.unidify_args is None


def test_idify_indir_outdir(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    in_dir = tmp_path / "in"
    out_dir = tmp_path / "out"

    mock_context = MockBatchResultsContext()
    monkeypatch.setattr(
        nasty._cli,
        nasty._cli.BatchResults.__name__,  # type: ignore
        mock_context.MockBatchResults,
    )

    main("idify", "--in-dir", str(in_dir), "--out-dir", str(out_dir))

    assert mock_context.init_args == (in_dir,)
    assert mock_context.idify_args == (out_dir,)
    assert mock_context.unidify_args is None


def test_unidify_stdin(monkeypatch: MonkeyPatch, capsys: CaptureFixture) -> None:
    monkeypatch.setattr(
        sys,
        "stdin",
        StringIO(
            "\n".join(
                ("1115690002233556993", "1115690615612825601", "1115691710657499137")
            )
        ),
    )
    main("unidify")

    assert [
        "We encourage submissions of new, previously, or concurrently published "
        "research. The event should be a forum for researchers to exchange ideas, "
        "discuss work, and get feedback. We hope you'll consider submitting your work.",
        "We'll have talks from research leaders on the latest advances in NLP. "
        "@NandoDF will be giving the keynote and more speakers will be announced soon. "
        "https://t.co/SB3URxn6ab",
        "Registration will open soon. In the meantime, we'll hope you'll save the date "
        "and consider joining us for what should be a fun day of listening to "
        "stimulating talks, mingling with like-minded people, exchanging ideas, and "
        "maybe even striking up a collaboration.",
    ] == [
        Tweet(json.loads(line)).text
        for line in capsys.readouterr().out.strip().split("\n")
    ]


def test_unidify_indir(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    mock_context = MockBatchResultsContext()
    monkeypatch.setattr(
        nasty._cli,
        nasty._cli.BatchResults.__name__,  # type: ignore
        mock_context.MockBatchResults,
    )

    main("unidify", "--in-dir", str(tmp_path))

    assert mock_context.init_args == (tmp_path,)
    assert mock_context.idify_args is None
    assert mock_context.unidify_args == (tmp_path,)


def test_unidify_indir_outdir(monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
    in_dir = tmp_path / "in"
    out_dir = tmp_path / "out"

    mock_context = MockBatchResultsContext()
    monkeypatch.setattr(
        nasty._cli,
        nasty._cli.BatchResults.__name__,  # type: ignore
        mock_context.MockBatchResults,
    )

    main("unidify", "--in-dir", str(in_dir), "--out-dir", str(out_dir))

    assert mock_context.init_args == (in_dir,)
    assert mock_context.idify_args is None
    assert mock_context.unidify_args == (out_dir,)



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC
from typing import Dict, Mapping, Optional, Type, TypeVar, cast

from overrides import overrides
from typing_extensions import Final

from .._util.typing_ import checked_cast
from ..tweet.tweet import TweetId
from .request import DEFAULT_BATCH_SIZE, DEFAULT_MAX_TWEETS, Request

_T_ConversationRequest = TypeVar("_T_ConversationRequest", bound="ConversationRequest")


class ConversationRequest(Request, ABC):
    def __init__(
        self,
        tweet_id: TweetId,
        *,
        max_tweets: Optional[int] = DEFAULT_MAX_TWEETS,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ):
        super().__init__(max_tweets=max_tweets, batch_size=batch_size)
        self.tweet_id: Final = tweet_id

    @overrides
    def to_json(self) -> Mapping[str, object]:
        obj: Dict[str, object] = {
            "type": None,  # Will be set in super(), but forces order.
            "tweet_id": self.tweet_id,
        }
        obj.update(super().to_json())
        return obj

    @classmethod
    @overrides
    def from_json(
        cls: Type[_T_ConversationRequest], obj: Mapping[str, object]
    ) -> _T_ConversationRequest:
        assert obj["type"] == cls.__name__
        return cls(
            tweet_id=checked_cast(TweetId, obj["tweet_id"]),
            max_tweets=(
                cast(Optional[int], obj["max_tweets"])
                if "max_tweets" in obj
                else DEFAULT_MAX_TWEETS
            ),
            batch_size=(
                checked_cast(int, obj["batch_size"])
                if "batch_size" in obj
                else DEFAULT_BATCH_SIZE
            ),
        )



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__version__ = "dev"
try:
    from nasty._version import __version__  # type: ignore
except ImportError:
    pass

__version_info__ = tuple(
    (int(part) if part.isdigit() else part)
    for part in __version__.split(".", maxsplit=4)
)


import logging

from nasty.__main__ import main
from nasty.batch.batch import Batch
from nasty.batch.batch_entry import BatchEntry
from nasty.batch.batch_results import BatchResults
from nasty.request.conversation_request import ConversationRequest
from nasty.request.replies import Replies
from nasty.request.request import DEFAULT_BATCH_SIZE, DEFAULT_MAX_TWEETS, Request
from nasty.request.search import DEFAULT_FILTER, Search, SearchFilter
from nasty.request.thread import Thread
from nasty.tweet.conversation_tweet_stream import ConversationTweetStream
from nasty.tweet.tweet import Tweet, TweetId, User, UserId
from nasty.tweet.tweet_stream import TweetStream

__all__ = [
    "main",
    "Batch",
    "BatchEntry",
    "BatchResults",
    "ConversationRequest",
    "Replies",
    "DEFAULT_BATCH_SIZE",
    "DEFAULT_MAX_TWEETS",
    "Request",
    "DEFAULT_FILTER",
    "Search",
    "SearchFilter",
    "Thread",
    "ConversationTweetStream",
    "Tweet",
    "TweetId",
    "User",
    "UserId",
    "TweetStream",
]

# Don't show log messages in applications that don't configure logging.
# See https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger(__name__).addHandler(logging.NullHandler())



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Iterable, Optional, Sequence, Type, cast

from overrides import overrides
from typing_extensions import Final

from .._util.typing_ import checked_cast
from ..request.replies import Replies
from ..tweet.tweet import TweetId
from .conversation_retriever import ConversationRetriever, ConversationRetrieverBatch


class RepliesRetrieverBatch(ConversationRetrieverBatch):
    @overrides
    def _tweet_ids(self) -> Iterable[TweetId]:
        instructions: Final = cast(
            Sequence[Any], self._json["timeline"]["instructions"]
        )

        # Replies are nested in conversation threads contained in instructions. Batches
        # look like this:
        # {
        #     ...
        #     "timeline": {
        #         "id": "Conversation-1155486497451184128",
        #         "instructions": [
        #             {
        #                 "addEntries": {
        #                     "entries": [
        #                         {"entryId": "tweet-1155486497451184128", ...},
        #                         {"entryId": "conversationThread-11554883561...", ...},
        #                         {"entryId": "conversationThread-11566020334...", ...},
        #                         {"entryId": "conversationThread-11555057423...", ...},
        #                         ...
        #                         {"entryId": "cursor-bottom-8067885539403591668", ...},
        #                     ]
        #                 }
        #             },
        #             ...
        #         ],
        #         ...
        #     },
        # }
        #
        # The Tweet that corresponds to the requested ID has a "tweet-..." entry in the
        # first batch (this is missing from following batches), with multiple
        # "conversationThread-..." entries following and a "cursor-bottom-..." entry at
        # the end, which contains the contain the cursor needed to fetch the next batch.
        # Seldom it will be a "cursor-showMoreThreads-..." or a
        # "cursor-showMoreThreadsPrompt-..." reply. If no more replies exist the cursor
        # entry will also not exist.

        if "addEntries" not in instructions[0]:
            return

        for entry in instructions[0]["addEntries"]["entries"]:
            if entry["entryId"].startswith("tweet-"):
                # We do not want to return the Tweet with the requested ID because it is
                # not a reply to itself.
                pass

            elif entry["entryId"].startswith("conversationThread-"):
                # Conversation thread entries look like this:
                # {
                #     "entryId": "conversationThread-1155488356165398528",
                #     "sortIndex": "8067885539403591669",
                #     "content": {
                #         "timelineModule": {
                #             "items": [
                #                 {
                #                     "entryId": "tweet-1155488356165398528",
                #                     "item": {
                #                         "content": {
                #                             "tweet": {
                #                                 "id": "1155488356165398528",
                #                                 "displayType": "SelfThread",
                #                             }
                #                         }
                #                     },
                #                 },
                #                 {
                #                     "entryId": "tweet-1155490920621473792",
                #                     "item": {
                #                         "content": {
                #                             "tweet": {
                #                                 "id": "1155490920621473792",
                #                                 "displayType": "SelfThread",
                #                             }
                #                         }
                #                     },
                #                 },
                #                 ...
                #             ]
                #         }
                #     },
                # }
                # That is, a conversation is a list of Tweet entries. Here the first
                # entry is a direct reply to the requested Tweet and all following
                # entries are replies to the previous entry.
                reply_tweet = entry["content"]["timelineModule"]["items"][0]

                if "tombstone" in reply_tweet["item"]["content"]:
                    # Sometimes Tweets become unavailable over time (for instance
                    # because they were deleted). They sometimes still show up in
                    # results but are only designated with a tombstone. We skip those
                    # when returning results.
                    # {
                    #     "entryId": "tweet-1079406406644715520",
                    #     "item": {
                    #         "content": {
                    #             "tombstone": {
                    #                 "displayType": "Inline",
                    #                 "tombstoneInfo": {
                    #                     "text": "",
                    #                     "richText": {
                    #                         "text": "This Tweet is unavailable.",
                    #                         "entities": [],
                    #                         "rtl": False,
                    #                     },
                    #                 },
                    #                 "epitaph": "Suspended",
                    #             }
                    #         },
                    #         ...
                    #     },
                    # }
                    self.num_tombstones += 1

                else:
                    yield checked_cast(
                        TweetId, reply_tweet["item"]["content"]["tweet"]["id"]
                    )

            elif entry["entryId"].startswith("label-"):
                # Sometimes additional replies can be loaded in the UI via a "Load more
                # replies" button. These replies appear under a "More replies" label,
                # which is added by this entry.
                # {
                #     "entryId": "label-8127279332145703061",
                #     "sortIndex": "8127279332145703061",
                #     "content": {
                #         "item": {
                #             "content": {
                #                 "label": {
                #                     "text": "More replies",
                #                     "displayType": "InlineHeader",
                #                 }
                #             }
                #         }
                #     },
                # }
                pass
            elif entry["entryId"].startswith("cursor-bottom-"):
                pass
            elif entry["entryId"].startswith("cursor-showMoreThreads-"):
                pass
            elif entry["entryId"].startswith("cursor-showMoreThreadsPrompt-"):
                pass
            elif entry["entryId"] == "novel_coronavirus_message":
                pass
            else:
                raise RuntimeError(
                    "Unknown entry type in entry-ID: '{}'.".format(entry["entryId"])
                )

    @overrides
    def _next_cursor(self) -> Optional[str]:
        instructions: Final = cast(
            Sequence[Any], self._json["timeline"]["instructions"]
        )

        if "addEntries" not in instructions[0]:
            return None

        # See the documentation of _tweet_ids_in_batch() on where the cursor entry
        # occurs. It looks like this:
        # {
        #     "entryId": "cursor-bottom-8067885539403591668",
        #     "sortIndex": "8067885539403591668",
        #     "content": {
        #         "operation": {
        #             "cursor": {
        #                 "value": "LBkWgMC1tbaij4kgJQISAAA=",
        #                 "cursorType": "Bottom",
        #             }
        #         }
        #     },
        # }

        cursor_entry = instructions[0]["addEntries"]["entries"][-1]
        if cursor_entry["entryId"].startswith("cursor-bottom"):
            return checked_cast(
                str, cursor_entry["content"]["operation"]["cursor"]["value"]
            )

        # Seldom the cursor entry looks like this instead:
        # {
        #     "entryId": "cursor-showMoreThreads-8067885539403590108",
        #     "sortIndex": "8067885539403590108",
        #     "content": {
        #         "operation": {
        #             "cursor": {
        #                 "value": "LBn2nQGAwLW1tqKPiSCAgLfZ/NqJjSCAwKf...",
        #                 "cursorType": "ShowMoreThreads",
        #                 "displayTreatment": {"actionText": "Show more replies"},
        #             }
        #         }
        #     },
        # }
        # This is very likely related to Twitter's UI of Show more replies, but
        # I don't know what the difference to regular replies is.
        #
        # In even rarer cases, it looks like this:
        # { "entryId": "cursor-showMoreThreadsPrompt-8127279332145704315",
        #   "sortIndex": "8127279332145704315",
        #   "content": {
        #     "operation": {
        #       "cursor": {
        #         "value": "LBn2QICAqL24y8K2HoTArbHpyZO2HoCApqXKh5...",
        #         "cursorType": "ShowMoreThreadsPrompt",
        #         "displayTreatment": {
        #           "actionText": "Show",
        #           "labelText": "Show additional replies, including those "
        #                        "that may contain offensive content" }}}}}
        elif cursor_entry["entryId"].startswith(
            "cursor-showMoreThreads-"
        ) or cursor_entry["entryId"].startswith("cursor-showMoreThreadsPrompt-"):
            return checked_cast(
                str, cursor_entry["content"]["operation"]["cursor"]["value"]
            )

        return None


class RepliesRetriever(ConversationRetriever[Replies]):
    @classmethod
    @overrides
    def _retriever_batch_type(cls) -> Type[RepliesRetrieverBatch]:
        return RepliesRetrieverBatch



#
# Copyright 2019-2020 Lukas Schmelzeisen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from datetime import date, datetime
from http import HTTPStatus
from pathlib import Path
from typing import Sequence

import pytest
import responses
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch

from nasty._util.io_ import read_file, read_lines_file, write_file
from nasty._util.json_ import JsonSerializedException, read_json, write_json
from nasty._util.typing_ import checked_cast
from nasty.batch.batch import Batch
from nasty.batch.batch_entry import BatchEntry
from nasty.batch.batch_results import BatchResults
from nasty.request.replies import Replies
from nasty.request.request import Request
from nasty.request.search import Search, SearchFilter
from nasty.request.thread import Thread

REQUESTS: Sequence[Request] = [
    Search("q"),
    Search("q", filter_=SearchFilter.PHOTOS, lang="de"),
    Search("q", since=date(2009, 1, 20), until=date(2017, 1, 20)),
    Replies("332308211321425920"),
    Replies("332308211321425920", max_tweets=50),
    Thread("332308211321425920"),
    Thread("332308211321425920", batch_size=100),
]


def _make_json_serialized_exception() -> JsonSerializedException:
    # Collect exception with trace.
    try:
        raise ValueError("Test Error.")
    except ValueError as e:
        return JsonSerializedException.from_exception(e)


# -- test_json_conversion_* ------------------------------------------------------------


@pytest.mark.parametrize("request_", REQUESTS, ids=repr)
def test_json_conversion_request(request_: Request) -> None:
    batch_entry = BatchEntry(request_, id_="id", completed_at=None, exception=None)
    assert batch_entry == batch_entry.from_json(batch_entry.to_json())


def test_json_conversion_completed_at() -> None:
    batch_entry = BatchEntry(
        Search("q"), id_="id", completed_at=datetime.now(), exception=None
    )
    assert batch_entry == batch_entry.from_json(batch_entry.to_json())


def test_json_conversion_exception() -> None:
    batch_entry = BatchEntry(
        Search("q"),
        id_="id",
        completed_at=None,
        exception=_make_json_serialized_exception(),
    )
    assert batch_entry == batch_entry.from_json(batch_entry.to_json())


# -- test_dump_load_requests_* ---------------------------------------------------------


@pytest.mark.parametrize("request_", REQUESTS, ids=repr)
def test_dump_load_single(request_: Request, tmp_path: Path) -> None:
    batch_file = tmp_path / "batch.jsonl"

    batch = Batch()
    batch.append(request_)
    batch.dump(batch_file)

    lines = list(read_lines_file(batch_file))
    assert 1 == len(lines)
    assert 0 != len(lines[0])

    batch2 = Batch()
    batch2.load(batch_file)
    assert list(batch) == list(batch2)


@pytest.mark.parametrize("num_batch_entries", [10, 505, 1000], ids=repr)
def test_dump_load_multiple(num_batch_entries: int, tmp_path: Path) -> None:
    batch_file = tmp_path / "batch.jsonl"

    batch = Batch()
    for i in range(1, num_batch_entries + 1):
        batch.append(Search(str(i), max_tweets=i, batch_size=i))
    batch.dump(batch_file)

    lines = list(read_lines_file(batch_file))
    assert num_batch_entries == len(lines)
    for line in lines:
        assert 0 != len(line)

    batch2 = Batch()
    batch2.load(batch_file)
    assert list(batch) == list(batch2)


# -- test_execute_* --------------------------------------------------------------------


def _assert_results_dir_structure(
    results_dir: Path,
    batch_entries: Sequence[BatchEntry],
    *,
    allow_empty: bool = False,
) -> None:
    assert results_dir.exists()
    assert 0 != len(batch_entries)

    batch_results = BatchResults(results_dir)

    # Can't create sets, because BatchEntry is not hashable, thus compare maps.
    assert {batch_entry.id: batch_entry for batch_entry in batch_entries} == {
        batch_entry.id: batch_entry for batch_entry in batch_results
    }

    for batch_entry in batch_results:
        assert batch_entry.completed_at is not None
        assert batch_entry.completed_at < datetime.now()
        assert batch_entry.exception is None

        tweets = list(batch_results.tweets(batch_entry))
        if not allow_empty:
            assert 0 != len(tweets)

        assert checked_cast(int, batch_entry.request.max_tweets) >= len(tweets)
        for tweet in tweets:
            assert (
                checked_cast(Search, batch_entry.request).query.lower()
                in json.dumps(tweet.to_json()).lower()
            )


def test_execute_success(tmp_path: Path) -> None:
    batch = Batch()
    batch.append(Search("trump", max_tweets=50))
    batch.append(Search("hillary", max_tweets=50))
    batch.append(Search("obama", max_tweets=50))
    assert batch.execute(tmp_path)
    _assert_results_dir_structure(tmp_path, list(batch))


def test_execute_success_parallel(tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("NASTY_NUM_WORKERS", "4")
    batch = Batch()
    for i in range(16):
        batch.append(
            Search(
                "trump",
                since=date(2019, 1, i + 1),
                until=date(2019, 1, i + 2),
                max_tweets=50,
            )
        )
    assert batch.execute(tmp_path)
    _assert_results_dir_structure(tmp_path, list(batch))


def test_execute_success_empty(tmp_path: Path) -> None:
    # Random string that currently does not match any Tweet.
    unknown_word = "c9dde8b5451149e683d4f07e4c4348ef"
    batch = Batch()
    batch.append(Search(unknown_word))
    results = batch.execute(tmp_path)
    assert results
    assert not list(results.tweets(results[0]))
    _assert_results_dir_structure(tmp_path, list(batch), allow_empty=True)


def test_execute_retrying_after_exception(
    tmp_path: Path, caplog: LogCaptureFixture
) -> None:
    batch = Batch()
    batch.append(Search("trump", max_tweets=50))

    batch_entry = batch[0]
    exception = _make_json_serialized_exception()
    batch_entry.exception = exception
    meta_file = tmp_path / batch_entry.meta_file_name
    write_json(meta_file, batch_entry)
    batch_entry.exception = None
    meta_stat1 = meta_file.stat()

    caplog.clear()
    assert batch.execute(tmp_path)
    assert 1 == len(  # Assert that log says we are retrying and the previous exception.
        [
            record
            for record in caplog.records
            if "Retrying" in record.msg and str(exception) in record.msg
        ]
    )

    _assert_results_dir_structure(tmp_path, list(batch))
    meta_stat2 = meta_file.stat()
    assert meta_stat1.st_mtime_ns < meta_stat2.st_mtime_ns


def test_execute_stray_data_file(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    batch = Batch()
    batch.append(Search("trump", max_tweets=50))

    batch_entry = batch[0]
    data = "Just some stray data."
    data_file = tmp_path / batch_entry.data_file_name
    write_file(data_file, data)
    data_stat1 = data_file.stat()

    assert not batch.execute(tmp_path)

    # Assert exception is saved.
    assert batch_entry == read_json(tmp_path / batch_entry.meta_file_name, BatchEntry)
    assert batch_entry.exception is not None
    assert batch_entry.exception.type == "ValueError"
    batch_entry.exception = None

    # Assert that previous data file is not modified.
    data_stat2 = data_file.stat()
    assert data_stat1.st_mtime == data_stat2.st_mtime
    assert data == read_file(data_file)

    # Delete data file and verify that it works now.
    data_file.unlink()
    caplog.clear()
    assert batch.execute(tmp_path)
    assert 1 == len([record for record in caplog.records if "Retrying" in record.msg])
    _assert_results_dir_structure(tmp_path, list(batch))


def test_execute_skipping(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    batch_file = tmp_path / "batch.jsonl"
    results_dir = tmp_path / "out"

    # Execute request for the first time.
    batch = Batch()
    batch.append(Search("trump", max_tweets=50))
    batch.dump(batch_file)
    assert batch.execute(results_dir)
    _assert_results_dir_structure(results_dir, list(batch))

    batch_entry = batch[0]
    meta_file = results_dir / batch_entry.meta_file_name
    data_file = results_dir / batch_entry.data_file_name
    meta_stat1 = meta_file.stat()
    data_stat1 = data_file.stat()

    # Execute same request again (should be skipped).
    batch = Batch()  # Recreate from dumped batch file so that batch entry IDs match.
    batch.load(batch_file)
    caplog.clear()
    assert batch.execute(results_dir)
    assert 1 == len([record for record in caplog.records if "Skipping" in record.msg])
    _assert_results_dir_structure(results_dir, list(batch))
    meta_stat2 = meta_file.stat()
    data_stat2 = data_file.stat()

    # Verify that files were not modified.
    assert meta_stat1.st_mtime_ns == meta_stat2.st_mtime_ns
    assert data_stat1.st_mtime_ns == data_stat2.st_mtime_ns


@pytest.mark.requests_cache_disabled
@responses.activate
def test_execute_exception_internal_server_error(tmp_path: Path) -> None:
    # Simulate 500 Internal Server Error on first request to Twitter.
    responses.add(
        responses.GET,
        "https://mobile.twitter.com/robots.txt",
        body="Crawl-delay: 1",
    )
    responses.add(
        responses.GET,
        "https://mobile.twitter.com/search",
        match_querystring=False,
        status=HTTPStatus.INTERNAL_SERVER_ERROR.value,
    )

    batch = Batch()
    batch.append(Search("trump", max_tweets=50))
    assert not batch.execute(tmp_path)
    batch_entry = batch[0]
    assert batch_entry == read_json(tmp_path / batch_entry.meta_file_name, BatchEntry)
    assert batch_entry.exception is not None
    assert batch_entry.exception.type == "UnexpectedStatusCodeException"
